import math
import os
import tarfile
import tempfile

from grid.tar import get_dir_size_and_count
from grid.tar import get_split_size
from grid.tar import tar_path


class TestTar:
    @staticmethod
    def test_get_dir_size_and_count():
        sizes = [1024 * 512, 1024 * 1024 * 5]

        for size in sizes:
            with tempfile.TemporaryDirectory() as temp_dir:
                data = os.urandom(size)
                with open(os.path.join(temp_dir, "a"), "wb") as f:
                    f.write(data)
                with open(os.path.join(temp_dir, "b"), "wb") as f:
                    f.write(data)
                assert get_dir_size_and_count(temp_dir, "a") == (size, 1)

    @staticmethod
    def test_tar_path():
        with tempfile.TemporaryDirectory() as temp_dir:
            source_dir = os.path.join(temp_dir, "source")
            inner_dir = os.path.join(source_dir, "dir")
            os.makedirs(inner_dir)
            with open(os.path.join(source_dir, "f1"), 'w') as f:
                f.write("f1")

            with open(os.path.join(inner_dir, "f2"), 'w') as f:
                f.write("f2")

            # Test directory

            target_file = os.path.join(temp_dir, "target.tar.gz")
            results = tar_path(source_path=source_dir, target_file=target_file)
            assert results.before_size > 0
            assert results.after_size > 0

            verify_dir = os.path.join(temp_dir, "verify")
            os.makedirs(verify_dir)
            with tarfile.open(target_file) as target_tar:
                target_tar.extractall(verify_dir)

            assert os.path.exists(os.path.join(verify_dir, "f1"))
            assert os.path.exists(os.path.join(verify_dir, "dir", "f2"))

            # Test single file

            f2_path = os.path.join(inner_dir, "f2")

            target_file = os.path.join(temp_dir, "target_file.tar.gz")
            results = tar_path(source_path=f2_path, target_file=target_file)
            assert results.before_size > 0
            assert results.after_size > 0

            verify_dir = os.path.join(temp_dir, "verify_file")
            os.makedirs(verify_dir)
            with tarfile.open(target_file) as target_tar:
                target_tar.extractall(verify_dir)

            assert os.path.exists(os.path.join(verify_dir, "f2"))

            # Test single file (local)

            current_path = os.getcwd()
            try:
                os.chdir(inner_dir)

                f2_path = "f2"

                target_file = os.path.join(temp_dir, "target_file_local.tar.gz")
                results = tar_path(source_path=f2_path, target_file=target_file)
                assert results.before_size > 0
                assert results.after_size > 0

                verify_dir = os.path.join(temp_dir, "verify_file_local")
                os.makedirs(verify_dir)
                with tarfile.open(target_file) as target_tar:
                    target_tar.extractall(verify_dir)

                assert os.path.exists(os.path.join(verify_dir, "f2"))
            finally:
                os.chdir(current_path)

    @staticmethod
    def test_get_split_size():
        split_size = get_split_size(minimum_split_size=1024 * 1000 * 10, max_split_count=10000, total_size=200000000001)

        # We shouldn't go over the max split count
        assert math.ceil(200000000001 / split_size) <= 10000

        split_size = get_split_size(minimum_split_size=1024 * 1000 * 10,
                                    max_split_count=10000,
                                    total_size=1024 * 500 * 1000 * 10)

        assert split_size == 1024 * 1000 * 10

    @staticmethod
    def test_tar_path_no_compression():
        with tempfile.TemporaryDirectory() as temp_dir:
            source_dir = os.path.join(temp_dir, "source")
            inner_dir = os.path.join(source_dir, "dir")
            os.makedirs(inner_dir)
            with open(os.path.join(source_dir, "f1"), 'w') as f:
                f.write("f1")

            with open(os.path.join(inner_dir, "f2"), 'w') as f:
                f.write("f2")

            target_file = os.path.join(temp_dir, "target.tar.gz")
            tar_path(source_path=source_dir, target_file=target_file, compression=False)

            verify_dir = os.path.join(temp_dir, "verify")
            os.makedirs(verify_dir)
            with tarfile.open(target_file) as target_tar:
                target_tar.extractall(verify_dir)

            assert os.path.exists(os.path.join(verify_dir, "f1"))
            assert os.path.exists(os.path.join(verify_dir, "dir", "f2"))



from pathlib import Path

from grid.downloader import DownloadableObject
from grid.downloader import Downloader


def test_downloader_downloads_data(tmpdir):
    """Downloader().download() downloads data in nested directories."""
    # Google logo
    url = 'https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_272x92dp.png'
    directory_path = 'nested_0/nested_1'
    filename = 'test-filename.png'
    objects = [DownloadableObject(url=url, download_path=directory_path, filename=filename)]

    # Download file
    D = Downloader(downloadable_objects=objects, base_dir=tmpdir)
    D.download()

    # Test that file as been downloaded.
    assert Path(tmpdir).joinpath(directory_path).joinpath(filename).exists()



from dataclasses import dataclass
from dataclasses import field
from enum import Enum
import json
import logging
import math
import os
from pathlib import Path
import shutil
from threading import Lock
from typing import Dict, List, Optional
import uuid

from click.exceptions import ClickException
from dataclasses_json import dataclass_json
from gql import Client
from gql import gql
import requests
import yaspin

from grid.tar import get_split_size
from grid.tar import tar_path
from grid.uploader import S3Uploader
from grid.uploader import UploadProgressCallback


def check_source(source: str):
    """
    Check if the source is a valid source for datastore
    """
    if source.startswith("http"):
        supported_extensions = [".zip", ".tar.gz", ".tar"]

        valid_extension = any(source.endswith(extension) for extension in supported_extensions)

        if valid_extension is False:
            raise ClickException(f"""
                Grid currently only supports remote files with the
                following extensions: {supported_extensions}
                """)

        # Check if url is accessible
        status = requests.head(source, allow_redirects=True)
        if status.status_code != 200:
            raise ClickException(f"Source url {source} is not reachable")
        return

    if os.path.exists(source) is False:
        raise ClickException(f"Source local directory path {source} doesn't exist")


def create_datastore(name: str, credential_id: str, client: Client, remote_source_url: Optional[str] = None):
    """Create datastore in Grid"""

    # Create Grid datastore directly in Grid without uploading,
    # since Grid will handle extraction and creating a optimizted
    # datastore automatically.

    mutation = gql("""
    mutation (
        $name: String!
        $credentialId: String!
        $remoteSourceUrl: String
        ) {
        createDatastore (
            properties: {
                    name: $name
                    credentialId: $credentialId
                    remoteSourceUrl: $remoteSourceUrl
                }
        ) {
        success
        message
        datastoreId
        datastoreVersion
        }
    }
    """)

    params = {'name': name, 'credentialId': credential_id, 'remoteSourceUrl': remote_source_url}

    result = client.execute(mutation, variable_values=params)
    success = result['createDatastore']['success']
    message = result['createDatastore']['message']
    if not success:
        raise ValueError(f"Unable to create datastore: {message}")

    datastoreId = result['createDatastore']['datastoreId']
    datastoreVersion = int(result['createDatastore']['datastoreVersion'])

    return (datastoreId, datastoreVersion)


def create_datastore_session(name: str, source: str, credential_id: str, compression: bool, client: Client):
    """Creates datastore session based on source"""
    check_source(source)

    if source.startswith("http"):
        return DatastoreRemoteUpload(name=name, source_url=source, credential_id=credential_id, client=client)

    s = DatastoreUploadSession(name=name, source_dir=source, credential_id=credential_id, compression=compression)
    s.configure(client=client)
    return s


class InvalidDatastoreError(Exception):
    """Error that is non-recoverable when uploading datastore"""


class UnsupportedDatastoreImplementationError(Exception):
    """Error when trying to recover a non-supported datastore implementation"""


class DatastoreUploadSteps(str, Enum):
    CREATE_DATASTORE = "create_datastore"
    GET_PRESIGNED_URLS = "get_presigned_urls"
    COMPRESS_SOURCE_DIRECTORY = "compress_source_directory"
    UPLOAD_PARTS = "upload_parts"
    MARK_UPLOAD_COMPLETE = "mark_upload_complete"


class DatastoreRemoteUpload:
    """
    This class handles creating a datastore in Grid.

    Attributes
    ----------
    name: str
        Name of the datastore
    source_url: str
        Source data url
    credential_id: str
        Credential Id to map datastore to
    """
    def __init__(self, name: str, source_url: str, credential_id: str, client: Client):
        self.name = name
        self.source_url = source_url
        self.credential_id = credential_id
        self.client = client
        self.logger = logging.getLogger("grid.datastore")

    def upload(self):
        spinner = yaspin.yaspin(text=f'Creating datastore {self.name}', color="yellow")
        spinner.start()

        try:
            create_datastore(name=self.name,
                             credential_id=self.credential_id,
                             client=self.client,
                             remote_source_url=self.source_url)
        except Exception as e:
            spinner.text = "Failed to create datastore"
            spinner.fail("✘")
            raise e

        spinner.text = "Finished creating datastore"
        spinner.ok("✔")


@dataclass_json
@dataclass
class DatastoreUploadSession(UploadProgressCallback):
    """
    This class handles uploading datastore

    Attributes
    ----------
    name: str
        Name of the datastore
    source_dir: str
        Source directory to upload from
    credential_id: str
        Credential Id to use for uploading
    presigned_urls: Dict[int, str]
        Presigned urls retrieved from backend
    etags: Dict[int, str]
        Etags per part after uploading to cloud
    session_path: str
        Path to session files (data and state file)
    session_state_file: str
        Path to session state file
    """
    # This version is a datastore implementation version. If we change this
    # version means we no longer able to recover sessions that is not the same
    # version. This is assuming to be fine as datastore upload recovering
    # should be a optimization.
    DATASTORE_VERSION = 2
    # Path storing datastore session and compresed data
    grid_datastores_path = '.grid/datastores'

    # Datastore upload fields
    name: str
    source_dir: str
    credential_id: str

    version: int = 1
    id: str = ""
    upload_id: str = ""
    original_size: int = 0
    compressed_size: int = 0
    part_count: int = 0

    # Session state
    session_id: str = ""
    session_path: str = ""
    session_state_file: str = ""
    last_completed_step: DatastoreUploadSteps = None

    presigned_urls: Dict[int, str] = field(default_factory=dict)
    etags: Dict[int, str] = field(default_factory=dict)

    # Compression
    compression: bool = False

    def configure(self, client: Client, session_id: str = ""):
        self.client = client
        self.lock = Lock()
        self.logger = logging.getLogger("uploader.DatastoreUploadSession")
        self.spinner = None

    @staticmethod
    def recover_sessions() -> List['DatastoreUploadSession']:
        """
        Recover all upload sessions that wasn't completed, so
        user can resume them.

        Returns
        -------
        List[DatastoreUploadSession]
            List of resumable sessions
        """
        logger = logging.getLogger("DatastoreUploadSession")
        sessions = []
        session_path = str(Path.home().joinpath(DatastoreUploadSession.grid_datastores_path))
        session_dirs = os.listdir(session_path)
        for session_dir in session_dirs:
            session_dir = os.path.join(session_path, session_dir)
            try:
                session = DatastoreUploadSession.recover(session_dir)
                last_completed_step = session.last_completed_step
                if last_completed_step in \
                    [None, DatastoreUploadSteps.COMPRESS_SOURCE_DIRECTORY]:
                    # We choose not to resume sessions that hasn't finished getting
                    # presigned_urls, since the datastore version is only determined
                    # when that's called, and we don't know what version we're going
                    # to be resuming.
                    raise ValueError("Session not yet finished compression")

                sessions.append(session)
            except (ValueError, UnsupportedDatastoreImplementationError) as e:
                logger.warning(f"Removing incomplete session {session_dir}, reason: {e}")
                shutil.rmtree(session_dir, ignore_errors=True)

        return sessions

    @staticmethod
    def recover(session_dir: str) -> "DatastoreUploadSession":
        """
        Recover session if sessions exists
        """
        session_state_file = os.path.join(session_dir, "session.json")
        if not os.path.exists(session_state_file):
            raise ValueError("Session state file does not exist")

        with open(session_state_file, "r") as f:
            content = json.load(f)
            datastore_version = 0
            if "DATASTORE_VERSION" in content:
                datastore_version = content["DATASTORE_VERSION"]
            if datastore_version != DatastoreUploadSession.DATASTORE_VERSION:
                name = content["name"]
                version = content["version"]
                session_name = f"{name}-v{version}"
                raise UnsupportedDatastoreImplementationError(f"Incomplete datastore session {session_name} is no " +
                                                              "longer supported. Please restart this upload.")

            session = DatastoreUploadSession.from_dict(content)

        return session

    def _update_progress(self, text: str):
        """
        Update current progress

        Parameters
        ----------
        text: str
            Latest progress text
        """
        if self.spinner:
            self.spinner.text = text

    def _get_presigned_urls(self):
        """Gets presigned urls from backend"""
        if self.part_count <= 0:
            raise InvalidDatastoreError(f"Invalid part count calculated for datastore: {self.part_count}")

        self._update_progress("Requesting presigned URLs from Grid...")

        query = gql("""
        query GetMultiPartPresignedUrls (
            $path: String!,
            $datastoreId: ID!,
            $count: Int!
        ) {
            getMultiPartPresignedUrls (
                path: $path,
                datastoreId: $datastoreId,
                count: $count
            ) {
                uploadId
                presignedUrls {
                    url
                    part
                }
            }
        }
        """)
        params = {
            'datastoreId': self.id,
            'count': self.part_count,
            # TODO: This is going to change once we implement
            # a new tarring scheme, so hardcoding the path here for now.
            'path': "data.tar.gz"
        }

        result = self.client.execute(query, variable_values=params)

        result = result['getMultiPartPresignedUrls']
        presigned_urls = result['presignedUrls']
        self.upload_id = result['uploadId']
        presigned_map = {}
        for url in presigned_urls:
            presigned_map[int(url['part'])] = url['url']

        self.presigned_urls = presigned_map

    @property
    def target_file(self) -> str:
        """Get target compressed data file"""
        return os.path.join(self.session_path, "data.tar.gz")

    def _create_datastore(self):
        self._update_progress("Creating datastore in Grid...")
        datastore_id, datastore_version = create_datastore(name=self.name,
                                                           credential_id=self.credential_id,
                                                           client=self.client)
        self.id = datastore_id
        self.version = datastore_version

    def _compress_source_directory(self):
        if self.compression:
            self._update_progress(f"Packaging and compressing datastore {self.name}...")
        else:
            self._update_progress(f"Packaging datastore {self.name}...")

        tar_results = tar_path(source_path=self.source_dir, target_file=self.target_file, compression=self.compression)

        self.original_size = tar_results.before_size
        self.compressed_size = tar_results.after_size
        split_size = get_split_size(self.compressed_size)
        self.part_count = math.ceil(self.compressed_size / split_size)

    def _create_uploader(self, presigned_urls: Dict[int, str], already_uploaded_parts: List[int]):
        size = self.compressed_size
        if size == 0:
            # We added compressed size later, so if a session didn't have it serialized,
            # we will look it up.
            # This is used for the progress bar only and not affect functionality.
            size = os.stat(self.target_file).st_size

        split_size = get_split_size(size)

        return S3Uploader(source_file=self.target_file,
                          presigned_urls=presigned_urls,
                          already_uploaded_parts=already_uploaded_parts,
                          name=f"{self.name}-{self.version}",
                          split_size=split_size,
                          total_size=size,
                          progress_callback=self)

    def _upload_parts(self):
        self._update_progress(f"Uploading datastore {self.name} (v{self.version}) datastore" + " to S3...")

        # Split urls depending on whether they were already uploaded or not.
        # This can happen if we resume an upload session.
        # The already uploaded ones are useful to show an accurate progress percentage.
        unuploaded_presigned_urls = {}
        already_uploaded_parts = []
        for part, url in self.presigned_urls.items():
            if part not in self.etags:
                unuploaded_presigned_urls[part] = url
            else:
                already_uploaded_parts.append(part)

        uploader = self._create_uploader(presigned_urls=unuploaded_presigned_urls,
                                         already_uploaded_parts=already_uploaded_parts)

        # Stoping and starting the spinner, as the progress bar
        # in uploader will conflict with the spinner

        if self.spinner:
            self.spinner.stop()

        uploader.upload()

        if self.spinner:
            self.spinner.start()

    def _mark_upload_complete(self):
        self._update_progress("Completing multi-part uploads with Grid...")
        mutation = gql("""
        mutation (
            $datastoreId: ID!
            $uploadId: String!
            $parts: JSONString!
            $path: String!
            ) {
            completeMultipartDatastoreUpload (
                properties: {
                        datastoreId: $datastoreId
                        uploadId: $uploadId
                        parts: $parts
                        path: $path
                    }
            ) {
            success
            message
            }
        }
        """)

        params = {
            'datastoreId': self.id,
            'uploadId': self.upload_id,
            'parts': json.dumps(self.etags),
            'path': 'data.tar.gz'
        }

        result = self.client.execute(mutation, variable_values=params)
        success = result['completeMultipartDatastoreUpload']['success']
        message = result['completeMultipartDatastoreUpload']['message']
        if not success:
            raise ValueError(f"Unable to complete multi-part upload: {message}")

        self._update_progress("Completing datastore upload...")

        mutation = gql("""
        mutation (
            $datastoreId: ID!
            $sizeMib: Int!
            ) {
            completeDatastoreUpload (
                properties: {
                        datastoreId: $datastoreId
                        sizeMib: $sizeMib
                    }
            ) {
            success
            message
            }
        }
        """)

        params = {'datastoreId': self.id, 'sizeMib': math.ceil(self.original_size / (1024**2))}

        result = self.client.execute(mutation, variable_values=params)
        success = result['completeDatastoreUpload']['success']
        message = result['completeDatastoreUpload']['message']
        if not success:
            raise ValueError(f"Unable to complete datastore upload: {message}")

    def _create_session_file(self, session_id: str):
        """
        Create session state file so we can resume upload progress.
        """
        self.session_path = os.path.join(self.grid_datastores_path, session_id)
        self.session_path = str(Path.home().joinpath(self.session_path))
        Path.home().joinpath(self.session_path).mkdir(parents=True, exist_ok=True)
        self.session_file = os.path.join(self.session_path, "session.json")

    @property
    def session_name(self) -> str:
        return f"{self.name}-v{self.version}"

    def upload(self):
        """
        Upload completes the full datastore upload operation,
        and also records the progress of the upload, so
        it can be resumed later.
        """
        if self.session_id == "":
            self.session_id = str(uuid.uuid4())

        self._create_session_file(self.session_id)

        self.spinner = yaspin.yaspin(text=f'Uploading datastore {self.name} v{self.version}', color="yellow")
        self.spinner.start()

        steps = [(self._create_datastore, DatastoreUploadSteps.CREATE_DATASTORE),
                 (self._compress_source_directory, DatastoreUploadSteps.COMPRESS_SOURCE_DIRECTORY),
                 (self._get_presigned_urls, DatastoreUploadSteps.GET_PRESIGNED_URLS),
                 (self._upload_parts, DatastoreUploadSteps.UPLOAD_PARTS),
                 (self._mark_upload_complete, DatastoreUploadSteps.MARK_UPLOAD_COMPLETE)]

        current_step = 0
        if self.last_completed_step:
            for i, step in enumerate(steps):
                if step[1] == self.last_completed_step:
                    current_step = i + 1
                    break

            if current_step == 0:
                raise ValueError(f"Unsupported upload step: " + self.last_completed_step)

        try:
            while current_step < len(steps):
                func, step = steps[current_step]
                func()
                self.last_completed_step = step
                self._write_session()
                current_step += 1

            self._remove_session()

            self.spinner.text = "Finished uploading datastore."
            self.spinner.ok("✔")

        except InvalidDatastoreError as e:
            self.spinner.fail("✘")
            self.logger.error(f"""
            An non-recoverable error occured when uploading datastore
            {self.session_name}.

            Please try again or contact Grid for support
            """)
            raise e
        except (Exception, KeyboardInterrupt) as e:
            self.spinner.fail("✘")
            message = f"""
            Whoops, your datastore creation failed!

            To resume, run:

            grid datastores resume {self.session_name}
            """
            self.logger.error(message)
            raise e
        finally:
            self.spinner.stop()

    def upload_part_completed(self, part: int, etag: str):
        """
        Mark part uploaded

        Parameters
        ----------
        part: int
            Part number
        etag: str
            ETag returned
        """
        self.logger.debug(f"Part {part} finished uploading")
        with self.lock:
            self.etags[part] = etag
            self._write_session()

    def _write_session(self):
        """
        Writes the session state into session file
        """
        with open(self.session_file, "w") as f:
            data = self.to_dict()
            data["DATASTORE_VERSION"] = self.DATASTORE_VERSION
            json.dump(data, f)

    def _remove_session(self):
        shutil.rmtree(self.session_path, ignore_errors=True)



from datetime import datetime
from datetime import timedelta
from datetime import timezone
import os
from pathlib import Path
import unittest

import click
from dateutil.parser import parse as date_string_parse
import pytest

from grid.utilities import check_description_isnt_too_long
from grid.utilities import check_environment_variables
from grid.utilities import check_is_python_script
from grid.utilities import convert_utc_string_to_local_datetime
from grid.utilities import get_abs_time_difference
from grid.utilities import get_experiment_duration_string
from grid.utilities import get_graphql_url
from grid.utilities import get_param_values
from grid.utilities import install_autocomplete
from grid.utilities import introspect_module
from grid.utilities import string_format_timedelta


def test_check_is_python_script():
    class Context:
        params = {}

    script = 'script.py'
    assert check_is_python_script(Context, None, script) == script

    bad_script = 'script.cpp'
    with pytest.raises(click.BadParameter):
        check_is_python_script(None, None, bad_script)


def test_check_description_isnt_too_long():
    """
    _check_description_isnt_too_long() checks that description isn't
    too long
    """
    d = 'a' * 199
    assert d == check_description_isnt_too_long(None, None, d)

    d = 'a' * 201
    with pytest.raises(click.BadParameter):
        check_description_isnt_too_long(None, None, d)


def test_install_autocomplete(mocker, tmpdir):
    mocker.patch("pathlib.Path.home", return_value=Path(tmpdir))
    home = Path.home()
    (home / ".bashrc").write_text("")
    (home / ".zshrc").write_text("")
    install_autocomplete()
    assert (Path.home() / ".grid/autocomplete/complete.zsh").exists()
    assert (Path.home() / ".grid/autocomplete/complete.bash").exists()
    assert (home / ".bashrc").read_text().count("grid") == 2
    assert (home / ".zshrc").read_text().count("grid") == 2

    # make sure we are not writing more than once
    install_autocomplete()
    assert (home / ".bashrc").read_text().count("grid") == 2
    assert (home / ".zshrc").read_text().count("grid") == 2


class UtilitiesTestCase(unittest.TestCase):
    """Test utilities in credentials."""
    def test_convert_utc_string_to_local_datetime(self):
        date_string = "2016-08-29T16:02:54.884Z"
        date_date = convert_utc_string_to_local_datetime(date_string)
        assert isinstance(date_date, datetime)

    def test_get_abs_time_difference(self):
        date_time_now = datetime.now()
        date_a = date_time_now - timedelta(hours=1)
        date_b = date_time_now
        result = get_abs_time_difference(date_1=date_a, date_2=date_b)
        inverse_result = get_abs_time_difference(date_1=date_b, date_2=date_a)
        assert isinstance(result, timedelta)
        assert int(result.total_seconds()) == 60 * 60
        assert int(inverse_result.total_seconds()) == 60 * 60

    def test_conversion_with_time_difference_now(self):
        finished_at = datetime.now(timezone.utc)
        created_at = date_string_parse('2020-05-16T21:21:10.745898+00:00')
        delta = get_abs_time_difference(finished_at, created_at)
        assert isinstance(delta, timedelta)

    def test_string_formating_duration(self):
        a = datetime.now()
        b = datetime.now() - timedelta(hours=1)
        c = a - b

        result = string_format_timedelta(c)
        assert result == '0d-00:59:59'

    def test_environment_variables(self):
        var = '_TEST_VAR'

        #  Checks that the env var isn't set.
        assert not os.getenv(var)

        #  Tests that function raises error.
        with pytest.raises(ValueError):
            check_environment_variables(variables=[var])

        #  Tests that it doesn't raise an error
        #  if variable is set.
        os.environ[var] = 'foo'
        assert check_environment_variables(variables=[var])

    def test_introspect_module(self):
        #  Creates a class that represents the same
        #  module mechanics: it includes the __all__
        #  attribute that has a string reference to
        #  a class property (test).
        class M:
            @staticmethod
            def test():
                return True

            __all__ = ['test']

        #  Let's that the generator returns
        #  the defined method.
        for func in introspect_module(M):
            assert isinstance(func, staticmethod)

    def test_get_param_vals(self):
        command = "command.py --flag1 val1 --flag2 --flag3 val3 --flag4"
        assert get_param_values(command) == ['val1', "True", "val3", "True"]

    def test_get_experiment_duration_string(self):
        created_at = str(datetime.now(timezone.utc) - timedelta(hours=3))
        started_running_at = str(datetime.now(timezone.utc) - timedelta(hours=2))
        finished_at = str(datetime.now(timezone.utc) - timedelta(hours=1))
        experiment_queued_result = get_experiment_duration_string(created_at=created_at,
                                                                  started_running_at=None,
                                                                  finished_at=None)
        assert experiment_queued_result >= f"{0}d-{3:02d}:{0:02d}:{0:02d}"
        assert experiment_queued_result < f"{0}d-{3:02d}:{1:02d}:{0:02d}"
        experiment_running_result = get_experiment_duration_string(created_at=created_at,
                                                                   started_running_at=started_running_at,
                                                                   finished_at=None)
        assert experiment_running_result >= f"{0}d-{2:02d}:{0:02d}:{0:02d}"
        assert experiment_running_result < f"{0}d-{2:02d}:{1:02d}:{0:02d}"
        experiment_finished_result = get_experiment_duration_string(created_at=created_at,
                                                                    started_running_at=started_running_at,
                                                                    finished_at=finished_at)
        assert experiment_finished_result >= f"{0}d-{1:02d}:{0:02d}:{0:02d}"
        assert experiment_finished_result < f"{0}d-{1:02d}:{1:02d}:{0:02d}"

    @staticmethod
    def test_get_graphql_url():
        url_with_graphql = 'http://localhost:8000/graphql'
        url_without_graphql = 'http://localhost:8000'
        assert get_graphql_url(url_without_graphql) == 'http://localhost:8000/graphql'
        assert get_graphql_url(url_with_graphql) == 'http://localhost:8000/graphql'



from tests.mock_backend import GridAIBackenedTestServer

CLIENT = GridAIBackenedTestServer()


def test_check_url():
    query = """
    query CheckURL ($url: String!) {
        checkUrl (url: $url) {
            statusCode
        }
    }
    """
    values = {'url': 'foo'}
    result = CLIENT.execute(query, values)
    assert result.get('errors') is None
    assert result.get('checkUrl')['statusCode'] == 200



import click
import pytest

from grid.cli.utilities import validate_config
from grid.cli.utilities import validate_disk_size_callback


def test_validate_config_raises_exception():
    """validate_config() raises exception when disk size too small."""
    config = {'compute': {'train': {'disk_size': 10}}}
    with pytest.raises(click.ClickException):
        validate_config(cfg=config)


def test_validate_disk_size_callback_raises_exception():
    """validate_disk_size_callback() raises exception when value too low."""
    with pytest.raises(click.BadParameter):
        validate_disk_size_callback(None, None, 40)


def test_validate_disk_size_callback():
    """validate_disk_size_callback() returns value if correct."""
    expected = 200
    result = validate_disk_size_callback(None, None, expected)
    assert result == expected



from click.testing import CliRunner
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid
import grid.commands.credentials as credentials

RUNNER = CliRunner()


class TestCredentials:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        credentials.gql = lambda x: x

        create_test_credentials()

    def test_credentials_without_arguments_prints_table(self, monkeypatch):
        """grid credentials without arguments fails"""
        def mp_get_credentials(*args, **kwargs):
            return [{
                'credentialId': 'test-cred-0',
                'provider': 'aws',
                'alias': 'my credential name',
                'createdAt': resolvers.now,
                'lastUsedAt': resolvers.now,
                'defaultCredential': True
            }]

        monkeypatch.setattr(resolvers, 'get_user_credentials', mp_get_credentials)
        result = RUNNER.invoke(cli.credentials, [])
        assert result.exit_code == 0
        assert not result.exception
        assert 'test-cred-0' in result.output

    def test_sets_default_credential(self):
        """grid credentials --set_default CREDENTIAL_ID sets default credential"""
        credential_id = 'test-cred'
        result = RUNNER.invoke(cli.credentials, ['--set_default', credential_id])
        assert result.exit_code == 0
        assert not result.exception
        assert credential_id in result.output

    def test_add_credentials(self):
        """grid credentials add adds new credentials"""
        file_path = 'tests/data/aws.json'
        result = RUNNER.invoke(cli.credentials, ['add', '--provider', 'aws', '--file', file_path])
        assert result.exit_code == 0
        assert not result.exception
        assert 'created' in result.output

    def test_add_credentials_with_bad_provider_fails(self):
        """grid credentials with bad provider fails"""
        file_path = 'tests/data/aws.json'
        result = RUNNER.invoke(cli.credentials, ['add', '--provider', 'test-fails', '--file', file_path])
        assert result.exit_code == 2
        assert result.exception



import click
from gql.transport.exceptions import TransportQueryError
import pytest
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

import grid.client as grid
import grid.observables.interactive_node as interactive_node_observable


def mp_execute(*args, **kwargs):
    raise Exception("{'message':'not found'}")


class TestIneteractiveObservable:
    @classmethod
    def setup_class(cls):
        create_test_credentials()

        grid.Grid._init_client = monkey_patch_client
        interactive_node_observable.gql = lambda x: x

        cls.grid = grid.Grid(load_local_credentials=False)
        cls.grid._init_client()

    def test_get(self, capsys):
        interactive_node_observable.env.SHOW_PROCESS_STATUS_DETAILS = False
        interactive_node_observable.env.DEBUG = True
        obsevable = interactive_node_observable.InteractiveNode(client=self.grid.client)

        obsevable.get()
        captured = capsys.readouterr()

        expected_columns = ['Name', 'Status']
        for column in expected_columns:
            assert column in captured.out

    def tets_get_handles_error(self, monkeypatch):
        def mp_transport_error(*args, **kwargs):
            raise TransportQueryError('test')

        monkeypatch.setattr(self.grid.client, 'execute', mp_transport_error)
        obsevable = interactive_node_observable.InteractiveNode(client=self.grid.client)
        with pytest.raises(click.ClickException):
            obsevable.get()



import ast
from datetime import datetime
from datetime import timezone

import click
from dateutil.parser import parse as date_string_parse
from gql import Client
from gql import gql
from gql.transport.exceptions import TransportQueryError
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TextColumn

import grid.globals as env
from grid.observables.base import BaseObservable
from grid.observables.base import TASK_CLASS_IGNORE
from grid.observables.base import TASK_CLASS_MAPPING
from grid.utilities import get_abs_time_difference
from grid.utilities import string_format_timedelta


class InteractiveNode(BaseObservable):
    """
    Base observable for a Grid interactive node.

    Parameters
    ----------
    client: Client
        GQL client
    """
    def __init__(self, client: Client):
        self.client = client

        super().__init__(client=client, spinner_load_type="Interactive Nodes")

    # TODO: This isn't currently used anywhere. It may be useful to also
    # show pre-run status for tasks related to creating an interactive node.
    def _get_task_status(self, interactive_node_id: str):  # pragma: no cover
        """
        Gets task status for a given interactive node.

        Parameters
        ----------
        interactive_node_id: str
            Interactive node ID.
        """
        self.spinner.start()

        query = gql("""
        query GetInteractiveTaskStatus ($interactiveNodeId: ID!) {
            getInteractiveTaskStatus(interactiveNodeId: $interactiveNodeId) {
                status
                message
                dependencies {
                    taskId
                    status
                    taskType
                    message
                    error
                }
            }
        }
        """)
        params = {'interactiveNodeId': interactive_node_id}

        try:
            result = self.client.execute(query, variable_values=params)
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            self.spinner.fail("✘")
            raise click.ClickException(message)

        # Gets dependenceis.
        dependencies = result['getRunTaskStatus']['dependencies']

        #  Figure out the status of each task.
        dependency_data = {}
        for task in dependencies:
            if task['taskType'] in TASK_CLASS_IGNORE:
                continue

            dependency_data[task['taskType']] = {
                'status': task['status'],
                'error': task['error'],
                'message': task['message'],
                'status_message': TASK_CLASS_MAPPING[task['taskType']]
            }

        #  Display the status of setting-up the cluster & environment
        #  to the user.
        self.spinner.text = 'Done!'
        self.spinner.ok("✔")

        #  Show message if there's an error in a pre-start task.
        all_status = [s['status'] for s in dependency_data.values()]
        if any(s == 'failed' for s in all_status) and \
            not env.SHOW_PROCESS_STATUS_DETAILS:
            click.echo(f'''
        The Interactive Node "{interactive_node_id}" failed to start due to a setup error.
        You can see the errors by running

            grid status {interactive_node_id} --details

            ''')

        #  Create progress bar.
        click.echo('\n')
        with Progress(TextColumn("[bold blue]{task.description}", justify="left"), BarColumn(bar_width=55),
                      "[self.progress.percentage]{task.percentage:>3.1f}%") as progress_bar:

            #  Render progress bar based on finished tasks.
            task_id = progress_bar.add_task('Creating Interactive Node', start=True, total=len(all_status))

            progress = sum(1 for s in all_status if s == 'finished')
            progress_bar.update(task_id, advance=progress)

        #  If there's an error with creating a cluster, then
        #  print that error to the terminal.
        if env.SHOW_PROCESS_STATUS_DETAILS:
            for key, value in dependency_data.items():
                if value.get('error'):
                    styled_key = click.style(TASK_CLASS_MAPPING[key], fg='red')
                    error = value.get('error')
                    for line in error.splitlines():
                        click.echo(f'[{styled_key}] {line}')
                else:
                    styled_key = click.style(TASK_CLASS_MAPPING[key], fg='green')
                    message = value.get('message')
                    click.echo(f'[{styled_key}] {message} ... ')

    def get(self):
        self.spinner.start()
        query = gql("""
        query GetInteractiveNodes{
            getInteractiveNodes {
                interactiveNodeId
                interactiveNodeName
                clusterId
                createdAt
                jupyterUrl
                status
                config {
                    diskSize
                    instanceType
                }
            }
        }
        """)

        result = {}

        try:
            result = self.client.execute(query)

            if not result['getInteractiveNodes']:
                self.spinner.fail("✘")
                if env.DEBUG:
                    click.echo(result['getInteractiveNodes'])
            else:
                self.spinner.ok("✔")

        #  Raise any other errors that the backend may raise.
        except TransportQueryError:  # skipcq: PYL-W0703
            self.spinner.fail("✘")

        # Create table with results.
        table_rows = 0
        table_cols = ['Name', 'Status', 'Instance Type', 'Duration', 'URL']
        table = BaseObservable.create_table(columns=table_cols)

        if result and result['getInteractiveNodes']:
            for row in result['getInteractiveNodes']:
                url = row['jupyterUrl']

                # Calculate how long the interactive node has been up.
                created_at = date_string_parse(row['createdAt'])
                delta = get_abs_time_difference(datetime.now(timezone.utc), created_at)
                duration_str = string_format_timedelta(delta)

                # Add rows to table.
                if row['status'] == 'finished':
                    status = 'ready'
                else:
                    status = row['status']

                # Replace anything that doesn't have data with
                # a placeholder '-'.
                all_rows = [row['interactiveNodeName'], status, row['config']['instanceType'], duration_str, url]
                all_rows = ['-' if not r else r for r in all_rows]
                table.add_row(*all_rows)

                # Adds iterator
                table_rows += 1

        #  Close the spinner.
        self.spinner.stop()

        # If there are no nodes active, add a
        # placeholder row indicating that.
        if not table_rows:
            table.add_row("None Active.", *[" " for i in range(len(table_cols) - 1)])

        # Print table
        self.console.print(table)

    def follow(self):  # pragma: no cover
        pass



from collections import OrderedDict
import random
import shlex
import sys
from typing import Any, Dict, Optional

import click
from coolname import generate_slug

from grid.cli.grid_credentials import print_creds_table
from grid.cli.utilities import read_config
from grid.cli.utilities import validate_config
from grid.cli.utilities import validate_disk_size_callback
from grid.client import Grid
import grid.globals as env
from grid.tracking import Segment
from grid.tracking import TrackingEvents
from grid.types import WorkflowType
from grid.utilities import check_description_isnt_too_long


def get_credentials(client: Grid, grid_credential: Optional[str] = None) -> Dict[str, str]:
    """
    Get credentials based on either grid credential or default credential

    Returns
    -------
    Dict[str, str]
        Credential auth
    """
    _default_credential = None
    creds = client.get_credentials()['getUserCredentials']
    if grid_credential:
        if creds:
            for credential in creds:
                if credential['credentialId'] == grid_credential:
                    return credential

        raise click.ClickException(f'Credential {grid_credential} does not exist. ' +
                                   'Use grid credentials to see available credential IDs.')

    _url = client.url.replace('/graphql', '')
    no_credentials_message = f"""
    No cloud credentials available! Visit {_url}/#/settings to
    add new cloud credentials.
    """
    if not creds:
        raise click.ClickException(no_credentials_message)

    if len(creds) == 1:
        cred = creds[0]
        return cred

    for credential in creds:
        if credential['defaultCredential']:
            click.echo(f"Using default cloud credentials {credential['credentialId']} "
                       f"to run on {credential['provider'].upper()}.")
            return credential

    # If no default credentials are available, raise an exception.
    m = """
    Detected multiple credentials. Which would you like to use?
    """
    print_creds_table(creds)
    raise click.ClickException(m)


def _check_run_name_is_valid(_ctx, _param, value):
    """Click callback that checks if a Run contains reserved names."""
    if value is not None:
        fail = False

        #  Check if the input is alphanumeric.
        _run_name = value.replace('-', '')
        if not _run_name.isalnum():
            fail = True

        #  Check if the allowed `-` character is not used
        #  at the end of the string.
        elif value.endswith('-') or value.startswith('-'):
            fail = True

        #  Check that the run name does not contain any
        #  uppercase characters.
        elif any(x.isupper() for x in value):
            fail = True

        if fail:
            raise click.BadParameter(f"""

            Invalid Run name: {value} the Run name must be lower case
            alphanumeric characters or '-', start with an alphabetic
            character, and end with an alphanumeric character (e.g. 'my-name',
            or 'abc-123').

                """)

    return value


def _check_ignore_warnings_flag(ctx, _param, value):
    """
    Click callback that assigns the value of the ignore warnigns callback
    to a global variable.
    """
    if value is not None:
        env.IGNORE_WARNINGS = value

    return value


def _print_submission_message(instance_type: str, provider: str, credential_id: str, entrypoint: str, grid_name: str,
                              datastore_name: str, datastore_version: int, datastore_mount_dir: str):
    """Prints train submission message to the client."""
    message = f"""
        Run submitted!
        `grid status` to list all runs
        `grid status {grid_name}` to see all experiments for this run

        ----------------------
        Submission summary
        ----------------------
        script:                  {entrypoint}
        instance_type:           {instance_type}
        cloud_provider:          {provider}
        cloud_credentials:       {credential_id}
        grid_name:               {grid_name}
        datastore_name:          {datastore_name}
        datastore_version:       {datastore_version}
        datastore_mount_dir:     {datastore_mount_dir}
        """
    click.echo(message)


def _generate_default_grid_config(client: Grid,
                                  grid_credential: str,
                                  instance_type: str,
                                  grid_strategy: Optional[str] = None,
                                  grid_trials: Optional[int] = None,
                                  gpus: Optional[int] = None,
                                  max_nodes: Optional[int] = None,
                                  disk_size: Optional[int] = None,
                                  memory: str = None,
                                  cpus: Optional[int] = None,
                                  processes: Optional[int] = None,
                                  datastore_name: Optional[str] = None,
                                  datastore_version: Optional[int] = None,
                                  datastore_mount_dir: Optional[str] = None,
                                  framework: Optional[str] = None):
    """
    Generates a new default config file for user if user hasn't
    submitted one.
    """
    _default_credential = get_credentials(client=client, grid_credential=grid_credential)

    #  Relevant defaults for user.
    #  TODO: Pull from user profile when we have
    #  that information collected.
    defaults = {'region': 'us-east-1'}

    _grid_config = {
        'compute': {
            'provider': {
                'vendor': _default_credential['provider'],
                'credentials': _default_credential['credentialId'],
                'region': defaults['region']
            },
            'train': {
                'framework': framework,
                'datastore_name': datastore_name,
                'datastore_version': datastore_version,
                'datastore_mount_dir': datastore_mount_dir,
                'instance': instance_type,
                'nodes': 0,  # How many nodes are created at first.
                'max_nodes': max_nodes,
                'gpus': gpus,
                'scale_down_seconds': 30 * 60,  # 30 minutes before scaling down.
                'disk_size': disk_size,
                'memory': memory,
                'cpus': cpus,
                'processes': processes
            }
        }
    }

    #  We add a `hyper_params` key to the default config
    #  if the user has passed the `grid_strategy` argument.
    if grid_strategy:
        if not _grid_config.get('hyper_params'):
            _grid_config['hyper_params'] = {}

        _grid_config['hyper_params'] = {'settings': {'strategy': grid_strategy}, 'params': {}}

        #  We also add the key trials if that has been
        #  passed. trials is only valid when doing a `random_search`
        #  at the moment. Our parser will not be bothered by its
        #  presence if we pass it by default.
        if grid_trials:
            _grid_config['hyper_params']['settings']['trials'] = grid_trials

    if env.DEBUG:
        click.echo('Grid Config used:')
        click.echo(_grid_config)

    return _grid_config


def _check_is_valid_extension(ctx, _param, value):
    """Click callback that checks if a file is a Python script."""
    if value is not None:
        if not any(value.endswith(e) for e in ['.py', '.sh']):
            message = 'You must provide a valid script (python or shell). ' \
                      'No script detected.'
            raise click.BadParameter(message)

        ctx.params['entrypoint'] = value
        return value


def _aws_node_to_nickname():
    aws_node_to_nicknames = OrderedDict({
        'p3.16xlarge': '8_v100_16gb',
        'p3dn.24xlarge': '8_v100_32gb',
        'g4dn.metal': '8_t4_16gb',
        'p2.8xlarge': '8_k80_12gb',
        'p3.8xlarge': '4_v100_16gb',
        'g4dn.12xlarge': '4_t4_16gb',
        'g3.16xlarge': '4_m60_8gb',
        'g3.8xlarge': '2_m60_8gb',
        'p3.2xlarge': '1_v100_16gb',
        'g4dn.8xlarge': '1_t4_16gb',
        'g4dn.4xlarge': '1_t4_16gb',
        'g4dn.2xlarge': '1_t4_16gb',
        'g4dn.xlarge': '1_t4_16gb',
        'g4dn.16xlarge': '1_t4_16gb',
        'p2.xlarge': '1_k80_12gb',
        'g3s.xlarge': '1_m60_8gb',
        'g3.4xlarge': '1_m60_8gb',
        't2.large': '2_cpu_8gb',
        't2.medium': '2_cpu_4gb'
    })
    return aws_node_to_nicknames


def _nickname_to_aws_nodes():
    aws_node_to_nickname = _aws_node_to_nickname()
    aws_nickname_to_node = {v: k for k, v in aws_node_to_nickname.items()}
    return aws_nickname_to_node


def _resolve_instance_type_nickname(ctx, _param, value):
    """
    Enables instance type shortcuts like:
    2_cpu_4gb for t2.large
    """
    nickname = value.lower()

    aws_nickname_to_node = _nickname_to_aws_nodes()

    # validate potential options for the node name
    possible_values = list(aws_nickname_to_node.keys()) + list(aws_nickname_to_node.values())
    if nickname not in possible_values:
        possible_options = '\n'.join(list(aws_nickname_to_node.keys()))
        click.BadParameter(f'{nickname} is not an available instance_type\n try one of these:\n{possible_options}')

    instance_type = nickname

    # if the value has _ then it's a nickname
    if '_' in nickname:
        instance_type = aws_nickname_to_node[nickname]
    return instance_type


def _get_instance_types(ctx, args, incomplete):
    # TODO: these should be retrieved from backend
    return list(_aws_node_to_nickname().keys())


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.option('--grid_config',
              '--g_config',
              'config',
              type=click.File('r'),
              required=False,
              callback=read_config,
              help='Path to Grid config YML')
@click.option('--grid_name',
              '--g_name',
              'name',
              type=str,
              required=False,
              help='Name for this run',
              callback=_check_run_name_is_valid)
@click.option('--grid_description',
              '--g_description',
              'description',
              type=str,
              required=False,
              help='Run description; useful for note-keeping',
              callback=check_description_isnt_too_long)
@click.option('--grid_credential', '--g_credential', 'credential', type=str, required=False, help='Cloud to run on')
@click.option('--grid_strategy',
              '--g_strategy',
              'strategy',
              type=click.Choice(['grid_search', 'random_search'], case_sensitive=False),
              required=False,
              help='Hyperparameter search strategy')
@click.option('--grid_trials',
              '--g_trials',
              'trials',
              type=int,
              required=False,
              help='Number of trials to run hyper parameter search')
@click.option('--grid_instance_type',
              '--g_instance_type',
              'instance_type',
              type=str,
              default='2_m60_8gb',
              callback=_resolve_instance_type_nickname,
              help='Instance type to start training session in',
              autocompletion=_get_instance_types)
@click.option('--grid_gpus',
              '--g_gpus',
              'gpus',
              type=int,
              required=False,
              default=0,
              help='Number of GPUs to allocate per experiment')
@click.option(
    '--grid_cpus',
    '--g_cpus',
    'cpus',
    type=int,
    required=False,
    default=0,  # Default will be set to 1 after alerting user
    help='Number of CPUs to allocate per experiment')
# TODO: Re-enable this when we want to enable mult-nodex
#@click.option(
#    '--grid_processes',
#    '--g_processes',
#    'processes',
#    type=int,
#    required=False,
#    default=1,
#    help='Number of training processes to run per experiment. Using larger than 1 ' \
#         'will start a multi-node training')
@click.option('--grid_disk_size',
              '--g_disk_size',
              'disk_size',
              type=int,
              required=False,
              default=200,
              callback=validate_disk_size_callback,
              help='The disk size in GB to allocate to each node in the cluster')
@click.option('--grid_max_nodes',
              '--g_max_nodes',
              'max_nodes',
              type=int,
              required=False,
              default=10,
              help='The maximum nodes to scale cluster to')
@click.option('--grid_memory',
              '--g_memory',
              'memory',
              type=str,
              required=False,
              help='How much memory an experiment needs')
@click.option('--grid_datastore_name',
              '--g_datastore_name',
              'datastore_name',
              type=str,
              required=False,
              help='Datastore name to be mounted in training')
@click.option('--grid_datastore_version',
              '--g_datastore_version',
              'datastore_version',
              type=str,
              required=False,
              help='Datastore version to be mounted in training')
@click.option('--grid_datastore_mount_dir',
              '--g_datastore_mount_dir',
              'datastore_mount_dir',
              type=str,
              required=False,
              default='/opt/datastore',
              help='Directory to mount Datastore in training job')
@click.option('--grid_framework',
              '--g_framework',
              'framework',
              type=str,
              required=False,
              default='lightning',
              help='Framework to use in training')
@click.option('--ignore_warnings',
              is_flag=True,
              required=False,
              help='If we should ignore warning when executing commands',
              callback=_check_ignore_warnings_flag)
@click.argument('script', type=click.Path(exists=True), callback=_check_is_valid_extension)
@click.argument('script_args', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def train(ctx, script: str, entrypoint: str, config: Optional[Dict], name: Optional[str], credential: Optional[str],
          strategy: Optional[str], trials: Optional[int], instance_type: str, gpus: Optional[int],
          disk_size: Optional[int], max_nodes: Optional[int], description: Optional[str], ignore_warnings: bool,
          script_args: Optional[Any], memory: Optional[str], cpus: Optional[int], datastore_name: Optional[str],
          datastore_version: Optional[str], datastore_mount_dir: Optional[str], framework: Optional[str]) -> None:
    """Trains scripts in remote infrastructure"""

    # Captures all other user arguments that are not
    # parsed automatically.
    script_arguments = list(script_args)

    # Captures invocation command exactly as it was
    # typed in the terminal. pipe.quote maintains original
    # quotation used in the CLI. We'll replace absolute
    # grid executable path with the `grid` alias.
    sys.argv[0] = "grid"
    invocation_command = " ".join(map(shlex.quote, sys.argv))

    if env.DEBUG:
        click.echo(f"Script Arguments {script_arguments}")
        click.echo(f"Entrypoint Script: {entrypoint}")
        click.echo(f"Hyperparams Search Strategy: {strategy}")
        click.echo(f"GPUs Requested: {gpus}")

    client = Grid()

    tracker = Segment()
    tracker.send_event(event=TrackingEvents.RUN_CREATED, properties={'user_input': invocation_command})

    # make a fun random name when user does not pass in a name
    if name is None:
        name = f'{generate_slug(2)}-{random.randint(0, 1000)}'
        click.echo(f'No --grid_name passed, naming your run {name}')

    #  If the user has not passed a grid config file,
    #  then generate one with a set of default options.
    #  We'll add a default instance and the user's
    #  default credentials.
    if not config:
        _grid_config = _generate_default_grid_config(client=client,
                                                     grid_credential=credential,
                                                     grid_strategy=strategy,
                                                     grid_trials=trials,
                                                     instance_type=instance_type,
                                                     processes=1,
                                                     gpus=gpus,
                                                     max_nodes=max_nodes,
                                                     disk_size=disk_size,
                                                     memory=memory,
                                                     cpus=cpus,
                                                     datastore_name=datastore_name,
                                                     datastore_version=datastore_version,
                                                     datastore_mount_dir=datastore_mount_dir,
                                                     framework=framework)

    else:
        _grid_config = config

    validate_config(cfg=_grid_config)

    client.validate_datastore_version(grid_config=_grid_config)
    datastore_version = _grid_config["compute"]["train"].get("datastore_version")

    # Warn user if they have not specified a cpu or gpu number so cpus will be set to 1
    cpus = _grid_config['compute']['train'].get('cpus')
    gpus = _grid_config['compute']['train'].get('gpus')
    if not cpus and not gpus:
        notice_str = click.style('WARNING', fg='yellow')
        cpu_warning = f"{notice_str} Neither a CPU or GPU number was specified. 1 CPU will be " + \
                      "used as a default. To use N GPUs pass in '--grid_gpus N' flag."

        _grid_config['compute']['train']['cpus'] = 1
        # We will be showing this warning to users, but not
        # issuing a prompt.
        if not env.IGNORE_WARNINGS:
            # click.confirm(cpu_warning, abort=True)
            click.echo(cpu_warning)

    # TODO: Re-enable this when we want to enable mult-nodex
    #if processes and not entrypoint.endswith(".py") and processes > 1:
    #    raise click.ClickException(
    #        "Multiple processes is only supported for " + "python programs")

    #  Send to client.
    client.train(config=_grid_config,
                 kind=WorkflowType.SCRIPT,
                 run_name=name,
                 run_description=description,
                 entrypoint=entrypoint,
                 script_args=script_arguments,
                 invocation_command=invocation_command)

    #  Shows a friendly submission message to the
    #  user.
    _print_submission_message(instance_type=_grid_config['compute']['train']['instance'],
                              provider=_grid_config['compute']['provider']['vendor'],
                              credential_id=_grid_config['compute']['provider']['credentials'],
                              entrypoint=entrypoint,
                              grid_name=name,
                              datastore_name=_grid_config['compute']['train']['datastore_name'],
                              datastore_version=_grid_config['compute']['train']['datastore_version'],
                              datastore_mount_dir=_grid_config['compute']['train']['datastore_mount_dir'])



import os

from click.testing import CliRunner
from tests.mock_backend import GridAIBackenedTestServer

from grid import cli
from grid.cli.grid_login import click
import grid.client as grid
import grid.globals as env

CREDENTIAL_ID = 'test-credential'
CREDENTIAL_ID_DEFAULT = 'defaultCredential'
RUNNER = CliRunner()

CONTENT = "content"


# Monkey patches the getting of credentials
def monkey_patch_get_credentials(self):
    return {'getUserCredentials': [{'credentialId': CREDENTIAL_ID, 'provider': 'aws', 'defaultCredential': True}]}


def monkey_patch_get_credentials_multiple(self):
    return {
        'getUserCredentials': [{
            'credentialId': CREDENTIAL_ID,
            'provider': 'aws',
            'defaultCredential': True
        }, {
            'credentialId': "Another",
            'provider': 'aws',
            'defaultCredential': False
        }]
    }


def monkey_patch_get_credentials_multiple_no_default(self):
    return {
        'getUserCredentials': [{
            'credentialId': CREDENTIAL_ID,
            'provider': 'aws',
            'defaultCredential': False,
            'createdAt': '2020-11-11',
            'lastUsedAt': '2020-11-11',
            'alias': None
        }, {
            'credentialId': "Another",
            'provider': 'aws',
            'defaultCredential': False,
            'createdAt': '2020-11-11',
            'lastUsedAt': '2020-11-11',
            'alias': None
        }]
    }


# Monkey patches the check for tokens
def monkey_patch_token_check(self):
    return {'checkUserGithubToken': {'hasValidToken': True}}


#  Monkey patch the client.
def monkey_patch_no_creds(self):
    return {'getUserCredentials': []}


#  Monkey patch the client.
def monkey_patch_wrong_creds(self):
    return {'getUserCredentials': [{'credentialId': 'wrong-cred', 'provider': 'aws', 'defaultCredential': True}]}


def monkey_patch_github_accessible(self, _):
    return True


def monkey_patch_github_not_accessible(self, _):
    return False


#  Monkey patches the GraphQL client to read from a local schema.
def monkey_patch_client(self):
    self.client = GridAIBackenedTestServer()


def test_train_fails_with_no_arguments(monkeypatch):
    """grid train without arguments fails"""
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)

    result = RUNNER.invoke(cli.train, [])
    assert result.exit_code == 2
    assert result.exception


def test_train_fails_with_no_credentials(monkeypatch):
    """grid train fails if user has no credentials."""
    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_no_creds)
    monkeypatch.setattr(grid.Grid, '_set_local_credentials', lambda x: True)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)

    result = RUNNER.invoke(cli.train, ['--', 'tests/data/script.py'])

    assert result.exit_code == 1
    assert result.exception


def test_train_fails_if_credentials_dont_exist(monkeypatch):
    """grid train fails if credentials don't exist for user."""
    monkeypatch.setattr(grid.Grid, '_set_local_credentials', lambda x: True)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)

    result = RUNNER.invoke(cli.train,
                           ['--grid_credential', 'test-fail', '--grid_cpus', 1, '--', 'tests/data/script.py'])

    assert result.exit_code == 1
    assert result.exception


def test_train_fails_if_wrong_credential(monkeypatch):
    """grid train fails if wrong credentials for user."""
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_wrong_creds)

    result = RUNNER.invoke(cli.train, ['--grid_credential', CREDENTIAL_ID, '--', 'tests/data/script.py'])

    assert result.exit_code == 1
    assert result.exception
    assert f"Credential {CREDENTIAL_ID} does not exist" in result.output


def test_train_fails_with_multiple_non_default_credentials(monkeypatch):
    """grid train fails if multiple non default credentials."""
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials_multiple_no_default)

    result = RUNNER.invoke(cli.train, ['--', 'tests/data/script.py'])

    assert result.exit_code == 1
    assert result.exception
    assert 'Detected multiple credentials. Which would you like to use?' in result.output


def test_train_fails_if_config_not_found(tmp_path, monkeypatch):
    """grid train fails if config not found"""
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)

    result = RUNNER.invoke(cli.train,
                           ['--grid_credential', CREDENTIAL_ID, '--grid_config', 'no-file', 'tests/data/script.py'])

    assert result.exit_code == 2
    assert result.exception
    assert 'Could not open file' in result.output


def test_train_fails_with_bad_config(tmp_path, monkeypatch):
    """grid train fails with bad config"""
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)

    bad_yamls = [
        """
        """, """
        compute
        """, """
        -
        """, """
        - random
        """
    ]

    for i, content in enumerate(bad_yamls):
        bad_config = tmp_path / f"bad_config_{i}.yaml"
        bad_config.write_text(content)

        result = RUNNER.invoke(
            cli.train,
            ['--grid_credential', CREDENTIAL_ID, '--grid_config',
             bad_config.as_posix(), '--', 'tests/data/script.py'])

        assert result.exit_code == 2
        assert result.exception
        assert 'Could not load your YAML config file' in result.output


def test_train_fails_with_incomplete_config(tmp_path, monkeypatch):
    """grid train fails with incomplete config"""
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)

    bad_yamls = [
        """
        no_compute:
        """, """
        compute:
            no_train:
        """, """
        compute:
            train:
                random: e
        """
    ]

    for i, content in enumerate(bad_yamls):
        bad_config = tmp_path / f"bad_config_{i}.yaml"
        bad_config.write_text(content)

        result = RUNNER.invoke(cli.train, [
            '--grid_credential', CREDENTIAL_ID, '--grid_config',
            bad_config.as_posix(), '--ignore_warnings', 'tests/data/script.py'
        ])

        assert result.exit_code == 1
        assert result.exception
        assert isinstance(result.exception, KeyError)

    bad_yamls = [
        """
        compute:
        """, """
        compute:
            - random
        """, """
        compute:
            no_train
        """, """
        compute:
            - random: value
        """, """
        compute:
            train
        """, """
        compute:
            train:
        """, """
        compute:
            train:
                random
        """, """
        compute:
            train:
                - random
        """, """
        compute:
            train:
                - random: value
        """
    ]
    for i, content in enumerate(bad_yamls):
        bad_config = tmp_path / f"bad_config_{i}.yaml"
        bad_config.write_text(content)

        result = RUNNER.invoke(cli.train, [
            '--grid_credential', CREDENTIAL_ID, '--grid_config',
            bad_config.as_posix(), '--ignore_warnings', 'tests/data/script.py'
        ])

        assert result.exit_code == 1
        assert result.exception
        assert isinstance(result.exception, TypeError)


def test_train_fails_if_description_is_too_long(monkeypatch):
    """grid train fails if description is too long."""
    description = 201 * '-'  #  200 characters is the limit.
    result = RUNNER.invoke(cli.train, [
        '--grid_description', description, '--grid_credential', CREDENTIAL_ID, '--grid_cpus', 1, '--',
        'tests/data/script.py'
    ])
    assert result.exit_code == 2
    assert result.exception


def test_train_fails_if_run_name_fails_validation(monkeypatch):
    """grid train fails if run name does not match validation requirements."""
    run_name = "RUN"  # capital letters not allowed in run name
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)
    monkeypatch.setattr(grid.Grid, '_set_local_credentials', lambda x: True)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)

    result = RUNNER.invoke(cli.train, [
        '--grid_credential',
        CREDENTIAL_ID,
        '--grid_name',
        run_name,
        '--grid_cpus',
        1,
        '--',
        'tests/data/script.py',
    ])
    assert result.exit_code == 2
    assert result.exception


def test_train_default_cred(monkeypatch):
    """grid train succeeds and uses default creds"""
    monkeypatch.setattr(grid, 'gql', lambda x: x)
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials_multiple)
    monkeypatch.setattr(grid.Grid, '_check_github_repo_accessible', monkey_patch_github_accessible)

    result = RUNNER.invoke(cli.train, ['--ignore_warnings', 'tests/data/script.py'])

    assert result.exit_code == 0
    assert not result.exception
    assert f'Using default cloud credentials {CREDENTIAL_ID} to run on' in result.output


def test_train_fails_if_github_repo_not_accessible(monkeypatch):
    """grid train fails and redirects to the settings page """

    launch_url = None

    def mp_launch(url):
        nonlocal launch_url
        launch_url = url
        return True

    monkeypatch.setattr(grid, 'gql', lambda x: x)
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)
    monkeypatch.setattr(grid.Grid, '_check_github_repo_accessible', monkey_patch_github_not_accessible)
    monkeypatch.setattr(click, 'launch', mp_launch)
    result = RUNNER.invoke(cli.train, ['--ignore_warnings', 'tests/data/script.py'])

    assert result.exit_code == 1
    assert result.exception
    assert 'Please grant access on the settings page' in result.output
    assert launch_url == os.getenv('GRID_URL', env.DEFAULT_GRID_URL).replace("/graphql", "") + '/#/settings'


def test_train_fails_no_ignore_warnings(monkeypatch):
    """grid train fails if no ignore_warnings"""
    monkeypatch.setattr(grid, 'gql', lambda x: x)
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)
    monkeypatch.setattr(grid.Grid, '_set_local_credentials', lambda x: True)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)

    env.DEBUG = True
    result = RUNNER.invoke(cli.train, ['--grid_credential', CREDENTIAL_ID, 'tests/data/script.py'])

    #assert result.exit_code == 0
    #assert not result.exception
    assert '1 CPU will be used as a default.' in result.output


def test_train(monkeypatch):
    """grid train returns 0 exit code"""
    monkeypatch.setattr(grid, 'gql', lambda x: x)
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)
    monkeypatch.setattr(grid.Grid, '_set_local_credentials', lambda x: True)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)
    monkeypatch.setattr(grid.Grid, '_check_github_repo_accessible', monkey_patch_github_accessible)

    env.DEBUG = True
    description = 199 * '-'  # 200 characters is the limit.
    result = RUNNER.invoke(cli.train, [
        '--grid_credential', CREDENTIAL_ID, '--grid_description', description, '--grid_strategy', 'grid_search',
        '--grid_trials', "1", '--ignore_warnings', 'tests/data/script.py'
    ])
    assert result.exit_code == 0
    assert not result.exception


def test_train_with_nickname(monkeypatch):
    """grid train returns 0 exit code"""
    monkeypatch.setattr(grid, 'gql', lambda x: x)
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)
    monkeypatch.setattr(grid.Grid, '_set_local_credentials', lambda x: True)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)
    monkeypatch.setattr(grid.Grid, '_check_github_repo_accessible', monkey_patch_github_accessible)

    env.DEBUG = True
    description = 199 * '-'  # 200 characters is the limit.
    result = RUNNER.invoke(cli.train, [
        '--grid_credential', CREDENTIAL_ID, '--grid_description', description, '--grid_instance_type', '2_cpu_4gb',
        '--grid_strategy', 'grid_search', '--grid_trials', "1", '--ignore_warnings', 'tests/data/script.py'
    ])
    assert result.exit_code == 0
    assert not result.exception


def test_train_with_framework(monkeypatch):
    """grid train takes a famework parameters"""
    monkeypatch.setattr(grid, 'gql', lambda x: x)
    monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)

    monkeypatch.setattr(grid.Grid, 'get_credentials', monkey_patch_get_credentials)
    monkeypatch.setattr(grid.Grid, '_set_local_credentials', lambda x: True)
    monkeypatch.setattr(grid.Grid, '_check_user_github_token', monkey_patch_token_check)
    monkeypatch.setattr(grid.Grid, '_check_github_repo_accessible', monkey_patch_github_accessible)

    env.DEBUG = True
    result = RUNNER.invoke(cli.train, [
        '--grid_credential', CREDENTIAL_ID, '--grid_description', 'test', '--grid_strategy', 'grid_search',
        '--grid_trials', "1", '--grid_framework', 'lightning', '--ignore_warnings', 'tests/data/script.py'
    ])
    assert result.exit_code == 0
    assert not result.exception



import re

from grid.client import Grid


class GridObject:
    """
    Base object. Inherited by all user-facing object; providing
    common methods and the Grid client instance.
    """
    def __init__(self):
        self.client = Grid()

    def _update_meta(self) -> None:
        """Updates object attributes with metadata from backend."""
        if not self._data:
            return

        for k, v in self._data.items():
            setattr(self, self._camel_case_to_snake_case(k), v)

    def refresh(self):  # pragma: no cover
        """
        Refreshes object metadata. This makes a query to Grid to fetch the
        object's latest data.
        """
        raise NotImplementedError

    @staticmethod
    def _camel_case_to_snake_case(data: str) -> str:
        """
        Original implementation from:

            * https://www.geeksforgeeks.org/python-program-to-convert-camel-case-string-to-snake-case/
        """
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', data)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



from typing import Optional

import click

from grid.client import Grid
import grid.globals as env
from grid.types import ObservableType
from grid.utilities import is_experiment


def _check_is_experiment(ctx, _param, value):
    """Callback that checks if a value is an experiment."""
    if not value:
        return value

    if is_experiment(value):
        raise click.BadArgumentUsage(f"Must pass a Run name, not Experiment: {value}")

    return value


@click.command()
@click.argument('run', type=str, nargs=1, required=False, callback=_check_is_experiment)
@click.option('--follow', type=bool, help='Follows Run setup progress in terminal', is_flag=True)
@click.option('--details', type=bool, help='Shows Run submission details', is_flag=True)
@click.option('--export',
              type=click.Choice(['csv', 'json'], case_sensitive=False),
              help='Exports status output to supplied file type')
def status(run: Optional[str] = None,
           follow: Optional[bool] = False,
           details: Optional[bool] = False,
           export: Optional[str] = None) -> None:
    """Checks the status of Runs and Experiments."""
    client = Grid()

    # Setup global flag to show build errors with
    # Runs or Experiments. This is more effective than
    # passing the flag around through all the invocations.
    env.SHOW_PROCESS_STATUS_DETAILS = details

    # Users can see the statuses of Experiments
    # by selecting runs.
    kind = ObservableType.RUN
    observable = []
    if run:
        observable = [run]
        kind = ObservableType.EXPERIMENT

    client.status(kind=kind, identifiers=observable, follow=follow, export=export)

    # If we have a Run, then don't print the global
    # interactive nodes table.
    if not run:
        client.status(kind=ObservableType.INTERACTIVE, identifiers=observable, follow=follow, export=export)



import random
from typing import Any, Optional

import click
from coolname import generate_slug
import yaml

from grid.client import Grid
import grid.globals as env
from grid.utilities import check_is_python_script


@click.group(invoke_without_command=True)
@click.pass_context
def slurm(ctx) -> None:
    """Manages SLURM workflows in Grid"""
    return


def _print_submission_message(instance_type: str, provider: str, credential_id: str, entrypoint: str, grid_name: str):
    """Prints train submission message to the client."""
    message = f"""
        Run submitted!
        `grid slurm status` to list all runs
        `grid slurm status {grid_name}` to see all experiments for this run

        ----------------------
        Submission summary
        ----------------------
        script:                  {entrypoint}
        instance_type:           {instance_type}
        provider:                {provider}
        credentials:             {credential_id}
        grid_name:               {grid_name}
        """
    click.echo(message)


@slurm.command()
@click.argument('alias', required=False, type=str)
@click.pass_context
def get_token(ctx, alias: str) -> None:
    """Get's an auth token for registering a grid-daemon in a SLURM cluster."""
    client = Grid()
    client.get_slurm_auth_token()


@slurm.command()
@click.option('--run_name', required=False, type=str)
@click.pass_context
def cancel(ctx, run_name: str) -> None:
    """Cancels a slurm run."""


@slurm.command()
@click.option('--run_name', required=False, type=str)
@click.pass_context
def logs(ctx, run_name: str) -> None:
    """Prints the logs from a slurm job."""


@slurm.command()
@click.option('--cluster_alias',
              required=True,
              type=str,
              help='The alias assigned to the daemon running in your SLURM cluster.')
@click.option('--run_name', required=False, type=str, help='Name of a run to see a status for.')
@click.pass_context
def status(ctx, cluster_alias: str, run_name: str = None):
    """Get's status for jobs in slurm."""
    client = Grid()

    client.slurm_status(cluster_alias=cluster_alias, run_name=run_name)


@slurm.command(context_settings=dict(ignore_unknown_options=True))
@click.option('--grid_name', '--g_name', 'name', type=str, required=False, help='Name for this run')
@click.option('--cluster_alias',
              type=str,
              required=True,
              help='The alias assigned to the daemon running in your SLURM cluster.')
@click.option(
    '--ntasks',
    type=int,
    required=False,
    help=
    'This option advises the Slurm controller that job steps run within the allocation will launch a maximum of number tasks and to provide for sufficient resources.'
)
@click.option('--ntasks_per_node', type=int, required=False, help='Request that ntasks be invoked on each node.')
@click.option('--job_time', type=str, required=False, help='Set a limit on the total run time of the job allocation.')
@click.option('--mem_per_cpu', type=str, required=False, help='Minimum memory required per allocated CPU.')
@click.option('--gpus', type=int, required=False, help='Specify the total number of GPUs required for the job.')
@click.option(
    '--cpus',
    type=int,
    required=False,
    help='Advise the Slurm controller that ensuing job steps will require ncpus number of processors per task.')
@click.option('--nodes',
              type=int,
              required=False,
              help='Request that a minimum of minnodes nodes be allocated to this job.')
@click.argument('script', required=True, type=click.Path(exists=True), callback=check_is_python_script)
@click.argument('script_args', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def train(ctx, script: str, entrypoint: str, name: Optional[str], cluster_alias: str, ntasks: int, ntasks_per_node: int,
          job_time: str, mem_per_cpu: int, gpus: int, cpus: int, nodes: int, script_args: Optional[Any]) -> None:
    """Trains on a SLURM cluster"""
    #  Captures all other user arguments that are not
    #  parsed automatically.
    script_arguments = list(script_args)

    if env.DEBUG:
        click.echo(f"Script Arguments {script_arguments}")

    # make a fun random name when user does not pass in a name
    if name is None:
        name = f'{generate_slug(2)}-{random.randint(0, 1000)}'
        click.echo(f'No --grid_name passed, naming your run {name}')

    config = {
        "job_name": name,
        "cluster_alias": cluster_alias,
        "script_name": entrypoint,
        "script_args": script_arguments,
        "ntasks": ntasks,
        "ntasks_per_node": ntasks_per_node,
        "job_time": job_time,
        "mem_per_cpu": mem_per_cpu,
        "gpus": gpus,
        "cpus": cpus,
        "nodes": nodes
    }

    _config_yaml = yaml.dump(config)
    client = Grid()

    if not cpus and not gpus:
        notice_str = click.style('NOTICE', fg='yellow')
        cpu_warning = f"""
        {notice_str}

        Neither a CPU or GPU number was specified, so 1 CPU will be
        used as a default.

        Would you like to still continue running this job?

        """
        if not env.IGNORE_WARNINGS:
            click.confirm(cpu_warning, abort=True)

    #  Send to client.
    result = client.train_on_slurm(config_str=_config_yaml, run_name=name, script_args=script_arguments)
    if result['success']:
        #  Shows a friendly submission message to the
        #  user.
        _print_submission_message(instance_type=f"SLURM Cluster with alias {cluster_alias}",
                                  provider="Slurm",
                                  credential_id=cluster_alias,
                                  entrypoint=entrypoint,
                                  grid_name=name)
    else:
        message = result['message']
        raise click.ClickException(f"Unable to submit train task with error {message}")



"""
Uploader that uploads files into Cloud storage
"""

from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from typing import AnyStr, Dict, List

import requests
from requests.adapters import HTTPAdapter
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TextColumn
from urllib3.util.retry import Retry


class UploadProgressCallback:
    """
    This class provides a interface for notifying upload progress
    """
    @staticmethod
    def upload_part_completed(part: int, etag: str):
        return NotImplemented


class S3Uploader:
    """
    This class uploads a source file with presigned urls to S3

    Attributes
    ----------
    source_file: str
        Source file to upload
    presigned_urls: Dict[int, str]
        Presigned urls dictionary, with key as part number and values as urls
    already_uploaded_parts: List[int]
        List of parts that have already been uploaded
    workers: int
        Amount of workers to upload parts
    retries: int
        Amount of retries when requests encounters an error
    backoff_factor: float
        A backoff factor to apply between attempts after the second try
    total_size: int
        Size of all files to upload
    name: str
        Name of this upload to display progress
    progress_callback: UploadProgressCallback
        Callback for notifying upload progress
    """
    workers: int = 8
    retries: int = 10000
    backoff_factor: float = 1.

    progress = Progress(TextColumn("[bold blue]{task.description}", justify="left"), BarColumn(bar_width=None),
                        "[self.progress.percentage]{task.percentage:>3.1f}%")

    def __init__(self,
                 presigned_urls: Dict[int, str],
                 already_uploaded_parts: List[int],
                 source_file: str,
                 total_size: int,
                 split_size: int,
                 name: str,
                 progress_callback: UploadProgressCallback = None):
        self.presigned_urls = presigned_urls
        self.already_uploaded_parts = already_uploaded_parts
        self.source_file = source_file
        self.total_size = total_size
        self.split_size = split_size
        self.name = name
        self.progress_callback = progress_callback

    @staticmethod
    def upload_s3_data(url: str, data: bytes, retries: int, backoff_factor: float) -> str:
        """
        Send data to s3 url

        Parameters
        ----------
        url: str
            S3 url string to send data to
        data: bytes
             Bytes of data to send to S3
        retries: int
            Amount of retries
        backoff_factor: float
            Retry backoff factor

        Returns
        -------
        str
            ETag from response
        """
        s = requests.Session()
        retries = Retry(total=retries, backoff_factor=backoff_factor, status_forcelist=[500, 502, 503, 504])
        s.mount('http://', HTTPAdapter(max_retries=retries))
        response = s.put(url, data=data)
        if 'ETag' not in response.headers:
            raise ValueError(f"Unexpected response from S3, response: {response.content}")

        return response.headers['ETag']

    def _get_size(self, part_path: str, part: int) -> AnyStr:
        """
        Returns the size of a specific part.

        Parameters
        ----------
        part_path: str
            Path to part file
        part: int
            Part number
        Returns
        -------
        AnyStr
            Part size
        """
        with open(part_path, 'rb') as f:
            pos = self.split_size * (part - 1)
            f.seek(pos)
            data = f.read(self.split_size)

        return data

    def _progress_already_uploaded(self, task_id: str) -> None:
        """
        Moves the progress bar to account for previously
        uploaded parts so that it shows consistent and
        recoverable progress when resuming the upload.

        Parameters
        ----------
        task_id: str
            Task id for tracking progress
        """
        for part in self.already_uploaded_parts:
            data = self._get_size(part_path=self.source_file, part=part)

            self.progress.update(task_id, advance=len(data))

    def _upload_part(self, url: str, part_path: str, part: int, task_id: str) -> None:
        """
        Upload part of the data file with presigned url

        Parameters
        ----------
        url: str
            Presigned url to upload to S3
        part_path: str
            Path to part file
        part: int
            Part number
        task_id: str
            Task id for tracking progress
        """
        with open(part_path, 'rb') as f:
            pos = self.split_size * (part - 1)
            f.seek(pos)
            data = f.read(self.split_size)

        etag = self.upload_s3_data(url, data, self.retries, self.backoff_factor)
        self.progress.update(task_id, advance=len(data))
        if self.progress_callback:
            self.progress_callback.upload_part_completed(part=part, etag=etag)

    def upload(self) -> None:
        """
        Upload files from source dir into target path in S3
        """
        task_id = self.progress.add_task("upload", filename=self.name, total=self.total_size)
        with ThreadPoolExecutor(max_workers=self.workers) as pool:
            try:
                self.progress.start()
                futures = []

                if self.already_uploaded_parts:
                    f = pool.submit(self._progress_already_uploaded, task_id)
                    futures.append(f)

                    for future in as_completed(futures):
                        future.result()

                for part, url in self.presigned_urls.items():
                    f = pool.submit(self._upload_part, url, self.source_file, part, task_id)
                    futures.append(f)

                    for future in as_completed(futures):
                        future.result()
            finally:
                self.progress.stop()



import os

import click
import pytest

from grid.commands import dependencies


class DependencyManager:
    has_change = True
    write_spec = lambda self: "Mock warning from test!"  # NOQA
    get_missing = lambda self: []  # NOQA


def get_click_confirm(holder):
    def click_confirm(message, *args, **kwargs):
        holder.clear()
        holder.append(message)
        raise click.Abort()

    return click_confirm


def test_check_dependency_spec_if_missing(monkeypatch):
    monkeypatch.setattr(DependencyManager, "get_missing", lambda self: ["fake-package"])
    monkeypatch.setattr(dependencies, "PipManager", DependencyManager)
    output = []
    monkeypatch.setattr(dependencies.click, "confirm", get_click_confirm(output))
    obj = dependencies.DependencyMixin()
    with pytest.raises(click.exceptions.Abort):
        obj._check_dependency_listing()
    assert "fake-package" in " ".join(output)


def test_check_dependency_spec_user_choice(monkeypatch):
    monkeypatch.setattr(dependencies, 'execute_git_command', lambda *args, **kwargs: "")
    monkeypatch.setattr(os, "listdir", lambda p: ["requirements.txt"])
    monkeypatch.setattr(dependencies, 'DISABLE_ON_DEPENDENCY_FILE_EXIST', False)

    monkeypatch.setattr(dependencies, "PipManager", DependencyManager)

    # Exit training and generate dependency files
    monkeypatch.setattr(click, "prompt", lambda *args, **kwargs: "1")
    obj = dependencies.DependencyMixin()
    assert obj._check_dependency_listing() is True

    # Ignore and continue to train
    monkeypatch.setattr(click, "prompt", lambda *args, **kwargs: "2")
    assert dependencies.DependencyMixin()._check_dependency_listing() is False

    # Ignore and exit
    monkeypatch.setattr(click, "prompt", lambda *args, **kwargs: "3")
    with pytest.raises(click.exceptions.Abort):
        dependencies.DependencyMixin()._check_dependency_listing()

    # No change in requirement file
    DependencyManager.has_change = False
    assert dependencies.DependencyMixin()._check_dependency_listing() is False


def test_package_manager_selection(monkeypatch):
    monkeypatch.setattr(dependencies, 'execute_git_command', lambda *args, **kwargs: "")
    monkeypatch.setattr(dependencies, 'DISABLE_ON_DEPENDENCY_FILE_EXIST', False)

    def mock_init(self):
        self.get_missing = lambda: []

    monkeypatch.setattr(dependencies.PipManager, "__init__", mock_init)
    monkeypatch.setattr(dependencies.PipManager, "has_change", False)
    monkeypatch.setattr(dependencies.CondaManager, "__init__", mock_init)
    monkeypatch.setattr(dependencies.CondaManager, "has_change", False)

    with monkeypatch.context() as mp:
        mp.setattr(os, "listdir", lambda p: ["requirements.txt"])
        obj = dependencies.DependencyMixin()
        obj._check_dependency_listing()
        assert isinstance(obj._deps_manager, dependencies.PipManager)

    with monkeypatch.context() as mp:
        mp.setattr(os, "listdir", lambda p: ["environment.yml"])
        obj = dependencies.DependencyMixin()
        obj._check_dependency_listing()
        assert isinstance(obj._deps_manager, dependencies.CondaManager)

    with monkeypatch.context() as mp:
        mp.setattr(os, "listdir", lambda p: [])
        obj = dependencies.DependencyMixin()
        obj._check_dependency_listing()
        assert isinstance(obj._deps_manager, dependencies.PipManager)

    with monkeypatch.context() as mp:
        mp.setattr(os, "listdir", lambda p: [])
        mp.setenv("CONDA_DEFAULT_ENV", "fake_conda_env")
        obj = dependencies.DependencyMixin()
        obj._check_dependency_listing()
        assert isinstance(obj._deps_manager, dependencies.CondaManager)


def test_if_both_pipfile_and_condafile(monkeypatch):
    monkeypatch.setattr(os, "listdir", lambda p: ["environment.yml", "requirements.txt"])
    obj = dependencies.DependencyMixin()
    assert obj._check_dependency_listing() is False



import os
import tempfile

from grid.uploader import S3Uploader
from grid.uploader import UploadProgressCallback


def test_s3_uploader():
    with tempfile.TemporaryDirectory() as tmpdir:
        data = {"aa": os.urandom(256), "ab": os.urandom(256), "ac": os.urandom(256), "ad": os.urandom(256)}
        keys = list(data.keys())
        keys.sort()
        with open(os.path.join(tmpdir, "data.tar.gz"), "wb") as f:
            for k in keys:
                f.write(data[k])

        urls = {1: "http://grid.ai/1", 2: "http://grid.ai/2", 3: "http://grid.ai/3", 4: "http://grid.ai/4"}
        already_uploaded_parts = [5]

        reverse_keys = {"aa": 1, "ab": 2, "ac": 3, "ad": 4}

        def mock_upload_data(url, part_data, *args):
            found_key = None
            for k, d in data.items():
                if part_data == d:
                    found_key = k
                    break

            assert found_key
            part = reverse_keys[found_key]
            assert urls[part] == url
            return str(part)

        class MockProgressUpdater(UploadProgressCallback):
            def __init__(self):
                self.etags = {}

            def upload_part_completed(self, part: int, etag: str):
                self.etags[part] = etag

        updater = MockProgressUpdater()
        uploader = S3Uploader(source_file=os.path.join(tmpdir, "data.tar.gz"),
                              presigned_urls=urls,
                              already_uploaded_parts=already_uploaded_parts,
                              progress_callback=updater,
                              name="test",
                              total_size=256 * 4,
                              split_size=256)

        uploader.upload_s3_data = mock_upload_data

        uploader.upload()
        assert updater.etags == {1: "1", 2: "2", 3: "3", 4: "4"}



import uuid

from click.testing import CliRunner
from click.testing import Result
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid
import grid.commands.credentials as credentials
import grid.datastore as datastore
import grid.uploader as uploader

RUNNER = CliRunner()


class TestDatastores:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x
        datastore.gql = lambda x: x
        credentials.gql = lambda x: x
        cls.name = str(uuid.uuid4())[:8]

        create_test_credentials()

    @staticmethod
    def assert_exit(result: Result, exit_code: int = 0):
        assert result.exit_code == exit_code, \
         f"exit code is not zero, output: {result.output}"

    def test_create_without_grid_credential_works(self, monkeypatch):
        """grid datastores create without passing credentials works"""
        def mp_get_credentials(*args, **kwargs):
            return [{
                'credentialId': 'test-cred-0',
                'provider': 'aws',
                'alias': 'my credential name',
                'createdAt': resolvers.now,
                'lastUsedAt': resolvers.now,
                'defaultCredential': True
            }]

        def mp_upload(*args, **kwargs):
            return {'key': 'value'}

        monkeypatch.setattr(uploader.S3Uploader, 'upload', mp_upload)
        monkeypatch.setattr(resolvers, 'get_user_credentials', mp_get_credentials)
        result = RUNNER.invoke(cli.datastores, ['create', '--source', 'tests/data', '--name', self.name])
        self.assert_exit(result)
        assert not result.exception
        assert 'Finished uploading datastore' in result.output

        result = RUNNER.invoke(cli.datastores, ['create', '--source', 'tests/data', '--name', self.name])
        self.assert_exit(result)
        assert not result.exception
        assert 'Finished uploading datastore' in result.output

    def test_list_prints_table_of_datastores(self):
        """grid datastores list prints table of datstores"""
        result = RUNNER.invoke(cli.datastores, ['list'])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Credential Id' in result.output
        assert 'Name' in result.output
        assert 'Version' in result.output
        assert 'Size' in result.output
        assert 'Created' in result.output
        assert 'Status' in result.output
        assert 'test datastore' in result.output
        assert 'Fail' in result.output

    def test_deletes_a_datastore(self):
        """grid datastores delete deletes a datastore"""
        result = RUNNER.invoke(cli.datastores,
                               ['delete', '--name', self.name, '--version', '1', '--grid_credential', 'test-cred-0'])
        self.assert_exit(result)
        assert not result.exception

    def test_deletes_a_datastore_fails(self, monkeypatch):
        """grid datastores delete fails to delete a datastore"""
        def delete_datastore(*args, **kwargs):
            return {"success": False}

        monkeypatch.setattr(resolvers, 'delete_datastore', delete_datastore)
        result = RUNNER.invoke(cli.datastores,
                               ['delete', '--name', self.name, '--version', '1', '--grid_credential', 'test-cred-0'])
        self.assert_exit(result)
        assert not result.exception
        assert 'Failed to delete datastore' in result.output

    def test_deletes_a_datstore_fails_with_ex(self, monkeypatch):
        """grid datastores delete fails to delete a datastore with exception"""
        def delete_datastore(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'delete_datastore', delete_datastore)
        result = RUNNER.invoke(cli.datastores,
                               ['delete', '--name', self.name, '--version', '1', '--grid_credential', 'test-cred-0'])
        self.assert_exit(result=result, exit_code=1)
        assert result.exception
        assert 'Failed to delete datastore' in result.output



from click.testing import CliRunner
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid

RUNNER = CliRunner()


class TestMetrics:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x

        create_test_credentials()

    def test_metrics_without_arguments(self):
        """grid metrics without arguments fails"""
        result = RUNNER.invoke(cli.metrics, [])
        assert result.exit_code == 2
        assert result.exception

    def test_metrics_fails_with_ex(self, monkeypatch):
        """grid metrics fails with exception"""
        def get_experiment_scalar_logs(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'get_experiment_scalar_logs', get_experiment_scalar_logs)
        result = RUNNER.invoke(cli.metrics, ["test-run", "train_loss"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to make query' in result.output

    def test_experiment_wthout_metrics_raises_error(self, monkeypatch):
        """grid logs on experiment without metrics raises exception"""
        def get_experiment_scalar_logs(*args, **kwargs):
            return []

        monkeypatch.setattr(resolvers, 'get_experiment_scalar_logs', get_experiment_scalar_logs)
        result = RUNNER.invoke(cli.metrics, ["test-run", "train_loss"])
        assert result.exit_code == 1
        assert result.exception
        assert 'No metrics available' in result.output



import json
from typing import Dict, Optional

import click
from gql import gql
from rich.console import Console


class CredentialsMixin:
    """
    Mixin adding properties to the Grid client. This mixin adds
    all credential methods.
    """
    def __init__(self):
        self.console = Console()

    def get_credentials(self):
        query = gql("""
        query {
            getUserCredentials {
                credentialId
                provider
                alias
                createdAt
                lastUsedAt
                defaultCredential
            }
        }
        """)
        result = self.client.execute(query)
        return result

    def credentials_add(self, provider: str, credentials: Dict[str, str], alias: str, description: str) -> None:
        """Adds a new credential set to Grid."""
        mutation = gql("""
            mutation (
                $provider: String!
                $credentials: JSONString!
                $alias: String
                $description: String
                ) {
                createCloudCredential (
                    properties: {
                                provider: $provider
                                credentials: $credentials
                                alias: $alias
                                description: $description
                            }
                ) {
                success
                message
                }
            }
            """)
        params = {
            'provider': provider,
            'credentials': json.dumps(credentials),
            'alias': alias,
            'description': description
        }
        result = self.client.execute(mutation, variable_values=params)
        click.echo(result['createCloudCredential']['message'])

    def credentials_update(self,
                           credential_id: str,
                           alias: Optional[str] = None,
                           description: Optional[str] = None,
                           is_default: Optional[bool] = None):
        """Updates a credential set in the Grid backend."""
        mutation = gql("""
            mutation UpdateCredential (
                $credentialId: ID!
                $alias: String
                $description: String
                $isDefault: Boolean
            ) {
                updateCloudCredential (
                    credentialId: $credentialId
                    alias: $alias
                    description: $description
                    isDefault: $isDefault
                ) {
                    success
                    message
                }
            }
            """)

        params = {'credentialId': credential_id, 'alias': alias, 'description': description, 'isDefault': is_default}

        result = self.client.execute(mutation, variable_values=params)
        click.echo(result['updateCloudCredential']['message'])
        click.echo(f"Set default credentials to ID {credential_id}.")

    def credentials_delete(self):  # pragma: no cover
        pass



from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch

from click.testing import CliRunner

from grid import cli
from grid.cli.grid_artifacts import _download_artifacts
from grid.core import Artifact

RUNNER = CliRunner()


@patch("grid.cli.grid_artifacts.Experiment", autospec=True)
@patch("grid.cli.grid_artifacts.Run", autospec=True)
def test_artifacts_succeeds(experiment, run):
    """grid train without arguments fails"""
    experiment.artifacts = MagicMock()
    run.experiments = MagicMock()

    result = RUNNER.invoke(cli.artifacts, ["foo-bar", "foo-bar-exp0"])
    assert result.exit_code == 0
    assert not result.exception


def test_download_artifacts(tmpdir):
    artifact = Artifact(url="https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_272x92dp.png",
                        path="test-path/",
                        filename="google.png")

    # path to download
    file_path = tmpdir / Path(artifact.path) / Path(artifact.filename)

    # download and verify
    _download_artifacts(artifacts=[artifact], download_dir=tmpdir)
    assert Path(file_path).exists()



from click.testing import CliRunner
from tests.utilities import create_local_schema_client

from grid import cli
from grid.client import Grid

RUNNER = CliRunner()


def monkey_patch_client(self) -> None:
    """Monkey patches the GraphQL client to read from a local schema."""
    self.client = create_local_schema_client()


def monkey_patch_history(self, *args, **kwargs) -> None:
    """Monkey patches history() to avoid making a backend query."""
    return


def test_history_without_arguments_succeeds(monkeypatch):
    """grid history without arguments succeeds"""
    monkeypatch.setattr(Grid, '_init_client', monkey_patch_client)
    monkeypatch.setattr(Grid, 'history', monkey_patch_history)

    result = RUNNER.invoke(cli.history, [])
    assert result.exit_code == 0
    assert not result.exception



from pathlib import Path

from tests.utilities import create_local_schema_client

from grid.client import env
from grid.client import Grid

SETTINGS_PATH = 'tests/data/settings.json'
CREDENTIALS_PATH = 'tests/data/credentials.json'


def monkey_patch_client(self):
    self.client = create_local_schema_client()


def test_load_global_settings(monkeypatch):
    """Grid() loads global settings alters globals."""

    monkeypatch.setattr(Grid, '_init_client', monkey_patch_client)
    Grid.grid_settings_path = SETTINGS_PATH
    Grid(credential_path=CREDENTIALS_PATH, load_local_credentials=False)

    assert env.DEBUG == True


def test_global_settings_create_file_if_not_exists():
    """Grid() creates file if not exists."""

    new_path = 'tests/data/home/settings.json'
    Grid.grid_settings_path = new_path
    Grid(credential_path=CREDENTIALS_PATH, load_local_credentials=False)

    P = Path.home().joinpath(new_path)
    assert P.exists()

    # cleanup
    P.unlink()



from typing import Any, Dict

import click
import yaml

from grid.tracking import Segment
from grid.tracking import TrackingEvents

DISK_SIZE_ERROR_MSG = "Invalid disk size, should be greater than 100Gb"


def validate_config(cfg: Dict[str, Any]) -> None:
    """
    Validates Grid config.

    Parameters
    ----------
    cfg: Dict[str, Any]
        Dictionary representing a Grid config
    """
    disk_size = cfg['compute']['train']['disk_size']
    if disk_size is not None and disk_size < 100:
        raise click.ClickException(DISK_SIZE_ERROR_MSG)

    tracker = Segment()
    tracker.send_event(TrackingEvents.CONFIG_PARSED, properties={'config': cfg})


def read_config(ctx, param, value):
    """
    Click callback that reads value from the config file and fix
    the structure if it doesn't exist.
    """
    grid_config = value
    if grid_config:
        #  Loads the YML file as passed by the
        #  user.
        try:
            grid_config = yaml.safe_load(value.read())
            if not isinstance(grid_config, dict):
                raise Exception("Unexpected file structure")
        except Exception as e:
            raise click.BadParameter(f'Could not load your YAML config file: {e}')

        #  Adds required structure to the base
        #  YML file, if that structure isn't there.
        if 'compute' not in grid_config:
            grid_config['compute'] = {}
        if 'train' not in grid_config['compute']:
            grid_config['compute']['train'] = {}
    return grid_config


def validate_disk_size_callback(ctx, param, value: int) -> int:
    """
    Validates the disk size upon user input.

    Parameters
    ----------
    ctx
        Click context
    param
        Click parameter
    value: int

    Returns
    --------
    value: int
        Unmodified value if valid
    """
    if value < 100:
        raise click.BadParameter(DISK_SIZE_ERROR_MSG)

    return value



import click
from click.testing import CliRunner
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
from grid.client import Grid
import grid.client as grid
import grid.globals as env


class TestView:
    """Test case for the view command"""
    @classmethod
    def setup_class(cls):
        cls.runner = CliRunner()

        cls.credentials_path = 'tests/data/credentials.json'
        cls.test_run = 'manual-test-run'
        cls.test_exp = 'manual-test-run-exp0'

        # Setup the global DEBUG flag to True.
        env.DEBUG = True

        #  Monkey patches the GraphQL client to read from a local schema.
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x

        create_test_credentials()

    def monkey_patched_method(self, *args, **kwargs):
        return True

    def monkey_patch_grid_status(self, *args, **kwargs):
        result = {'getRuns': [{'name': self.test_run}]}
        return result

    def monkey_patch_grid_status_not_ready(self, *args, **kwargs):
        result = {'getRuns': [{'name': self.test_run}]}
        return result

    def monkey_patch_exception(self, *args, **kwargs):
        raise click.ClickException('test')

    def test_view_opens_browser(self, monkeypatch):
        """grid view RUN_ID opens browser."""
        monkeypatch.setattr(click, 'launch', self.monkey_patched_method)

        result = self.runner.invoke(cli.view, ["run", self.test_run])

        assert result.exit_code == 0
        assert 'http' in result.output
        assert 'run' in result.output
        assert self.test_run in result.output

        result = self.runner.invoke(cli.view, ["experiment", f'{self.test_run}-exp1'])
        assert result.exit_code == 0
        assert 'http' in result.output
        assert 'experiment' in result.output
        assert self.test_run in result.output

    def test_view_exception(self, monkeypatch):
        monkeypatch.setattr(click, 'launch', self.monkey_patched_method)

        old_status = Grid.status
        monkeypatch.setattr(Grid, 'status', self.monkey_patch_exception)

        # Set attribute again.
        monkeypatch.setattr(Grid, 'status', old_status)



from datetime import datetime
from datetime import timezone
from typing import Optional

import click
from dateutil.parser import parse as date_string_parse
from gql import Client
from gql import gql
from gql.transport.exceptions import TransportQueryError

import grid.globals as env
from grid.observables.base import BaseObservable
from grid.utilities import get_abs_time_difference
from grid.utilities import string_format_date
from grid.utilities import string_format_timedelta


class Run(BaseObservable):
    def __init__(self, client: Client, identifier: Optional[str] = []):
        self.client = client

        self.identifier = identifier

        super().__init__(client=client, spinner_load_type="Runs")

    def get(self):
        """
        Gets the run status; either for a single run or all runs for
        user.
        """
        self.spinner.start()
        self.spinner.text = 'Getting Run status ...'

        query = gql("""
        query (
            $runName: ID
        ) {
            getRuns (runName: $runName) {
                name
                createdAt
                experiments {
                    experimentId
                }
                nExperiments
                nFailed
                nCancelled
                nRunning
                nCompleted
                nQueued
                nPending
                projectId
            }
        }
        """)
        run_name = None
        if self.identifier:
            run_name = self.identifier

        params = {'runName': run_name}

        try:
            result = self.client.execute(query, variable_values=params)
            self.spinner.text = 'Done!'
        except TransportQueryError as e:
            self.spinner.fail("✘")
            self.spinner.stop()
            if env.DEBUG:
                click.echo(str(e))

            raise click.ClickException('Query to Grid failed. Try again in a few minutes.')

        table_cols = [
            'Run',
            'Project',
            'Status',
            'Duration',
            'Experiments',
            'Running',
            'Queued',
            'Completed',
            'Failed',
            'Stopped',
        ]
        table = BaseObservable.create_table(columns=table_cols)

        #  Whenever we don't have yet submitted experiments,
        table_rows = 0
        for row in result['getRuns']:
            status = None

            # we only have 3 statuses for runs
            # running (if something is running)
            is_running = row['nRunning'] is not None and row['nRunning'] > 0

            # We classify queued and pending into queued
            queued = row['nQueued']
            pending = row['nPending']
            n_queued = queued + pending

            # If anything is queued, the the status of the entire
            # run is queued. All other statuses are running in
            # all other conditions.
            if n_queued > 0:
                status = 'queued'

            # If you have anything running (and nothing queued)
            # then, mark the run as running.
            elif is_running:
                status = 'running'

            # If it doesn't match the conditions above, just
            # skip this row and add the row and put it in history.
            else:
                # don't render table because it should be in history
                continue

            #  Change the printed key from `None` to `-` if the
            #  no data exists for those keys.
            keys = ['nExperiments', 'nRunning', 'nQueued', 'nCompleted', 'nFailed', 'nCancelled']
            for key in keys:
                if row[key] is None:
                    row[key] = '-'

            # Calculate the duration column
            created_at = date_string_parse(row['createdAt'])
            delta = get_abs_time_difference(datetime.now(timezone.utc), created_at)
            duration_str = string_format_timedelta(delta)

            table.add_row(row['name'], row['projectId'], status, duration_str, str(row['nExperiments']),
                          str(row['nRunning']), str(n_queued), str(row['nCompleted']), str(row['nFailed']),
                          str(row['nCancelled']))

            #  Let's count how many rows have been added.
            table_rows += 1

        #  Close the spinner.
        self.spinner.ok("✔")
        self.spinner.stop()

        # If there are no Runs to render, add a
        # placeholder row indicating none are active.
        if table_rows == 0:
            table.add_row("None Active.", *[" " for i in range(len(table_cols) - 1)])

        self.console.print(table)

        #  Print useful message indicating that users can run
        #  grid history.
        history_runs = len(result['getRuns']) - table_rows
        if history_runs > 0:
            click.echo(f'\n{history_runs} Run(s) are not active. Use `grid history` ' 'to view your Run history.')

        return result

    def get_history(self):
        """
        Fetches a complete history of runs. This includes runs that
        are not currently active.
        """
        self.spinner.start()
        self.spinner.text = 'Getting Runs ...'

        query = gql("""
        query (
            $runName: ID
        ) {
            getRuns (runName: $runName) {
                name
                createdAt
                experiments {
                    experimentId
                }
                nExperiments
                nFailed
                nCancelled
                nRunning
                nCompleted
                nQueued
            }
        }
        """)
        run_id = None
        if self.identifier:
            run_id = self.identifier[0]

        params = {'runName': run_id}

        result = self.client.execute(query, variable_values=params)

        self.spinner.text = 'Done!'
        self.spinner.ok("✔")
        self.spinner.stop()

        table_cols = ['Run', 'Created At', 'Experiments', 'Failed', 'Stopped', 'Completed']
        table = BaseObservable.create_table(columns=table_cols)

        #  Whenever we don't have yet submitted experiments,
        table_rows = result['getRuns']
        for row in table_rows:
            keys = ['nExperiments', 'nFailed', 'nCancelled', 'nCompleted']
            for key in keys:
                if row[key] is None:
                    row[key] = '-'

            # check if it is running
            is_running = row['nRunning'] is not None and row['nRunning'] > 0

            # check if queued
            is_queued = row['nQueued'] is not None and row['nQueued'] > 0

            # history is everything else
            if is_queued or is_running:
                continue

            created_at = string_format_date(date_string_parse(row['createdAt']))

            table.add_row(row['name'], created_at, str(row['nExperiments']), str(row['nFailed']),
                          str(row['nCancelled']), str(row['nCompleted']))

        # Add placeholder row if no records are available.
        if not table_rows:
            table.add_row("No History.", *[" " for i in range(len(table_cols) - 1)])

        self.console.print(table)

        return result

    def follow(self):  # pragma: no cover
        pass



"""
Tracking module for Grid. Here we include logic
for tracking how users interact with the backend,
which helps us have a better picture of what the
complete user journey is across our product.

All of our tracking is done via the use of Segment
events. [1]

References
----------
[1] https://segment.com/docs/sources/server/python/
"""
from enum import Enum
import json
import os
import platform
import sys
import traceback
from typing import Dict
import uuid

import analytics
import click

from grid.client import Grid
import grid.globals as env
from grid.metadata import __version__


class TrackingEvents(Enum):
    EXCEPTION = 'EXCEPTION'
    CLI_STARTED = 'Cli started'
    CLI_FINISHED = 'Cli finished'
    RUN_CREATED = 'Run created'
    CONFIG_PARSED = 'Config parsed'
    INTERACTIVE_NODE_CREATED = 'Interactive node created'


class Segment:
    """Class for sending data to Grid's Segment analytics tracker."""
    client: analytics.Client

    def __init__(self, enabled=not env.GRID_DISABLE_TRACKING):
        """
        Initialize Segment tracking class

        Parameters
        ----------
        enabled: bool
            is the metric sending enabled
        """
        self.client = analytics.Client(
            write_key=env.SEGMENT_KEY,
            on_error=self.handle_error,
            send=enabled,
        )
        environment = {k: v for k, v in os.environ.items() if not self._is_env_sensitive(k)}
        self.client_kwargs = {
            'context': {
                'uname': platform.uname()._asdict(),
                'env': environment,
                'args': sys.argv,
                'cwd': os.getcwd(),
                'version': __version__,
                'invocation_id': str(uuid.uuid4()),
            },
        }
        self.client_kwargs['context']['json'] = json.dumps(self.client_kwargs['context'])
        user_id = None
        try:
            user_id = Grid().user_id
        except click.ClickException:
            pass

        anonymous_id = uuid.getnode()
        if user_id:
            self.client.alias(anonymous_id, user_id, **self.client_kwargs)
            self.client_kwargs['user_id'] = user_id
        else:
            self.client_kwargs['anonymous_id'] = anonymous_id

        self.client.identify(**self.client_kwargs)

    @staticmethod
    def _is_env_sensitive(key: str) -> bool:
        """
        Checks if the environment variable is sensitive for most common
        sensitive env vars

        Parameters
        ----------
        key: str
            environment key whose sensitivity we're checking

        Returns
        -------
            True if the env variable is sensitive, False otherwise


        """
        # covers GRID_API_KEY, AWS... etc
        for it in [
                "ACCESS",
                "KEY",
                "PASSWORD",
                "SECRET",
                "TOKEN",
        ]:
            if it in key.upper():
                return True
        return False

    @staticmethod
    def handle_error(error, items):  # pragma: no cover
        """
        Segment on_error callback
        """
        if env.DEBUG:
            env.logger.debug(f'Segment error detected: {error}')

    def send_event(self, event: TrackingEvents, properties: Dict = {}):
        """
        Send event happening to the Segment

        Parameters
        ----------
        event: TrackingEvents
            event which happened
        properties:
            Additional event properties
        """
        properties['json'] = json.dumps(properties)
        self.client.track(event=event.value, properties=properties, **self.client_kwargs)

    def report_exception(self, e: Exception):
        """
        report exception happening to the Segment backend
        Parameters
        ----------
        e: Exception

        """
        props = {
            'error': type(e).__name__,
            'stacktrace': traceback.format_stack(),
        }
        if hasattr(e, 'format_message'):
            props['message'] = e.format_message()
        elif hasattr(e, 'message'):
            props['message'] = e.message
        self.send_event(
            event=TrackingEvents.EXCEPTION,
            properties=props,
        )

    def flush(self):
        self.client.flush()



import json
import math
import os
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock

from click.exceptions import ClickException
import pytest

from grid.datastore import check_source
from grid.datastore import create_datastore_session
from grid.datastore import DatastoreRemoteUpload
from grid.datastore import DatastoreUploadSession
from grid.datastore import DatastoreUploadSteps
from grid.datastore import requests
from grid.datastore import UnsupportedDatastoreImplementationError
import grid.datastore as datastore
from grid.tar import TarResults


class TestDatastore:
    @classmethod
    def setup_class(cls):
        datastore.gql = lambda x: x

    @staticmethod
    def test_check_source():
        with pytest.raises(ClickException):
            check_source("http://www.grid.ai/data.7z")

        with pytest.raises(ClickException):
            check_source("/not/exists")

        check_source("/usr")

    @staticmethod
    def test_datastore_recover(monkeypatch):
        with TemporaryDirectory() as tempdir:
            session = DatastoreUploadSession(name="test", source_dir="test_dir", credential_id="cc-abcdef")
            session.last_completed_step = DatastoreUploadSteps.GET_PRESIGNED_URLS
            session.session_file = os.path.join(tempdir, "session.json")
            session.presigned_urls = {
                1: "http://www.grid.ai/1",
                2: "http://www.grid.ai/2",
                3: "http://www.grid.ai/3",
                4: "http://www.grid.ai/4"
            }
            session.etags = {2: "etag2", 3: "etag3"}
            session.part_count = 4
            session._write_session()
            recovered_session = DatastoreUploadSession.recover(tempdir)
            assert session.last_completed_step == recovered_session.last_completed_step
            assert session.presigned_urls == recovered_session.presigned_urls
            assert session.part_count == recovered_session.part_count
            assert session.etags == recovered_session.etags

    @staticmethod
    def test_datastore_outdated_Version(monkeypatch):
        with TemporaryDirectory() as tempdir:
            session = DatastoreUploadSession(name="test", source_dir="test_dir", credential_id="cc-abcdef")
            session.last_completed_step = DatastoreUploadSteps.GET_PRESIGNED_URLS
            session.session_file = os.path.join(tempdir, "session.json")
            session.presigned_urls = {
                1: "http://www.grid.ai/1",
                2: "http://www.grid.ai/2",
                3: "http://www.grid.ai/3",
                4: "http://www.grid.ai/4"
            }
            session.etags = {2: "etag2", 3: "etag3"}
            session.part_count = 4
            session.DATASTORE_VERSION = 0
            session._write_session()
            with pytest.raises(UnsupportedDatastoreImplementationError):
                DatastoreUploadSession.recover(tempdir)

    @staticmethod
    def test_datastore_steps(monkeypatch):
        session = DatastoreUploadSession(name="test", source_dir="test_dir", credential_id="cc-abcdef")

        class MockS3Uploader:
            def __init__(self, session):
                self.session = session

            def upload(self):
                self.session.upload_part_completed(1, "etag1")
                self.session.upload_part_completed(2, "etag2")

        class MockClient:
            def execute(self, query, variable_values):
                if "createDatastore" in query:
                    assert variable_values == {'credentialId': 'cc-abcdef', 'name': 'test', 'remoteSourceUrl': None}

                    return {
                        'createDatastore': {
                            'datastoreId': 'abcde',
                            'datastoreVersion': 1,
                            'success': True,
                            'message': ''
                        }
                    }

                if "GetMultiPartPresignedUrls" in query:
                    assert variable_values == {'datastoreId': 'abcde', 'count': 50, 'path': 'data.tar.gz'}

                    return {'getMultiPartPresignedUrls': {'presignedUrls': [], 'uploadId': "upload1"}}

                if "completeMultipartDatastoreUpload" in query:
                    assert variable_values == {
                        'datastoreId': 'abcde',
                        'uploadId': 'upload1',
                        'path': 'data.tar.gz',
                        'parts': json.dumps({
                            1: "etag1",
                            2: "etag2"
                        })
                    }

                    return {'completeMultipartDatastoreUpload': {'success': True, 'message': ''}}

                if "completeDatastoreUpload" in query:
                    assert variable_values == {
                        'sizeMib': math.ceil((1024 * 1000 * 1000) / (1024**2)),
                        'datastoreId': 'abcde'
                    }

                    return {'completeDatastoreUpload': {'success': True, 'message': ''}}

                raise ValueError(f"Unexpected query {query}")

        client = MockClient()
        session.configure(client=client)

        def tar_path(**kwargs):
            return TarResults(before_size=1024 * 1000 * 1000, after_size=1024 * 1000 * 1000)

        def create_uploader(presigned_urls, already_uploaded_parts):
            return MockS3Uploader(session)

        monkeypatch.setattr('grid.datastore.tar_path', tar_path)
        monkeypatch.setattr(session, '_create_uploader', create_uploader)

        session.upload()

    def test_remote_datastore_url(self, monkeypatch):
        def mock_head(*args, **kwargs):
            request_mock = MagicMock()
            request_mock.status_code = 200
            request_mock.headers = {'Content-Length': 1024 * 1024 * 10}
            return request_mock

        monkeypatch.setattr(requests, 'head', mock_head)

        class MockClient:
            def execute(self, query, variable_values):
                if "createDatastore" in query:
                    assert variable_values == {
                        'credentialId': 'cc-abcdef',
                        'name': 'test-remote',
                        'remoteSourceUrl': "http://www.grid.ai/data.zip"
                    }

                    return {
                        'createDatastore': {
                            'datastoreId': 'abcde',
                            'datastoreVersion': 1,
                            'success': True,
                            'message': ''
                        }
                    }

                raise ValueError(f"Unexpected query {query}")

        client = MockClient()
        remote_session = create_datastore_session(name="test-remote",
                                                  source="http://www.grid.ai/data.zip",
                                                  credential_id="cc-abcdef",
                                                  client=client,
                                                  compression=False)

        assert isinstance(remote_session, DatastoreRemoteUpload)
        remote_session.upload()



import os
from pathlib import Path

from tests.utilities import create_local_schema_client

from grid.cli.grid_interactive import _generate_default_interactive_config
from grid.client import Grid
from grid.types import WorkflowType


class TestGridInteractive:
    @classmethod
    def setup_class(cls):
        #  Monkey patches the GraphQL client to read from a local schema.
        def monkey_patch_client(self):
            self.client = create_local_schema_client()

        #  skipcq: PYL-W0212
        Grid._init_client = monkey_patch_client

        cls.creds_path = 'tests/data/credentials.json'
        cls.grid_header_keys = ['X-Grid-User', 'X-Grid-Key']

        cls.train_kwargs = {
            'config': 'test-config',
            'kind': WorkflowType.SCRIPT,
            'run_name': 'test-run',
            'run_description': 'test description',
            'entrypoint': 'test_file.py',
            'script_args': ['--learning_rate', '0.001']
        }

    def remove_env(self):
        #  Makes sure that the GRID_CREDENTIAL_PATH is not set
        if os.getenv('GRID_CREDENTIAL_PATH'):
            del os.environ['GRID_CREDENTIAL_PATH']

    def setup(self):
        self.remove_env()

    def teardown(self):
        self.remove_env()

        #  Removes test credentials added to home path.
        P = Path.home().joinpath(self.creds_path)
        if P.exists():
            P.unlink(missing_ok=True)

    def mock_get_credentials(self):
        return {'getUserCredentials': [{'credentialId': 'test', 'provider': 'test', 'defaultCredential': True}]}

    def test_grid_interactive_config(self):
        """Grid().train() executes a training operation correctly."""
        grid = Grid(credential_path=self.creds_path, load_local_credentials=False)
        grid.get_credentials = self.mock_get_credentials

        interactive_config = _generate_default_interactive_config(credential_id='test',
                                                                  client=grid,
                                                                  instance_type='test')

        assert interactive_config['compute']['provider']['vendor'] == 'test'
        assert interactive_config['compute']['provider']['credentials'] == 'test'



from click.testing import CliRunner
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid

RUNNER = CliRunner()


class TestCancel:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x

        create_test_credentials()

    def test_cancel_without_arguments(self):
        """grid cancel without arguments fails"""
        result = RUNNER.invoke(cli.cancel, [])
        assert result.exit_code == 0
        result.stdout_bytes.decode('utf-8').startswith('Usage')

    def test_cancel_experiment(self):
        """grid cancel experiments with single and multi-input"""
        experiments = [f'test-experiment-exp{i}' for i in range(5)]

        result = RUNNER.invoke(cli.cancel, ["experiment", experiments[0]])
        assert result.exit_code == 0
        assert not result.exception
        assert 'cancelled successfully' in result.output

        result = RUNNER.invoke(cli.cancel, ["experiment", *experiments])
        assert result.exit_code == 0
        assert not result.exception
        assert 'cancelled successfully' in result.output

    def test_cancel_runs(self):
        """grid cancel Runs with single and multi-input"""
        runs = [f'test-run-{i}' for i in range(5)]
        result = RUNNER.invoke(cli.cancel, ["run", runs[0]])
        assert result.exit_code == 0
        assert not result.exception
        assert 'cancelled successfully' in result.output

        result = RUNNER.invoke(cli.cancel, ["run", *runs])
        assert result.exit_code == 0
        assert not result.exception
        assert 'cancelled successfully' in result.output

    def test_cancel_experiment_fails(self, monkeypatch):
        def mp_cancel_experiment(*args, **kwargs):
            return {'success': False}

        monkeypatch.setattr(resolvers, 'cancel_experiment', mp_cancel_experiment)
        result = RUNNER.invoke(cli.cancel, ["experiment", "some-experiment"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Cancel failed' in result.output

    def test_cancel_run_fails(self, monkeypatch):
        def mp_cancel_run(*args, **kwargs):
            return {'success': False}

        monkeypatch.setattr(resolvers, 'cancel_run', mp_cancel_run)
        result = RUNNER.invoke(cli.cancel, ["run", "some-run"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Cancel failed' in result.output

    def test_cancel_experiment_fails_when_experiment_not_found(self, monkeypatch):
        def mp_get_experiment_id(*args, **kwargs):
            return {"success": False}

        monkeypatch.setattr(resolvers, 'get_experiment_id', mp_get_experiment_id)
        result = RUNNER.invoke(cli.cancel, ["experiment", "some-experiment"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Cancel failed' in result.output
        assert 'Could not find experiment' in result.output

    def test_cancel_experiment_fails_when_experiment_not_found_exception(self, monkeypatch):
        def mp_get_experiment_id(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'get_experiment_id', mp_get_experiment_id)
        result = RUNNER.invoke(cli.cancel, ["experiment", "some-experiment"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Cancel failed' in result.output
        assert 'Could not find experiment' in result.output



import click
from click.exceptions import ClickException
import pytest

import grid.commands.checks as checks


def monkey_patch_click(*args, **kwargs):
    pass


def test_check_if_uncommited_files(monkeypatch):
    """
    Grid._check_if_uncommited_files() shows prompts if uncommited
    files are found
    """
    def monkey_pach_git_command(*args, **kwargs):
        return 'test_0.txt: needs update\ntest_1.txt: needs update'

    monkeypatch.setattr(checks, 'execute_git_command', monkey_pach_git_command)
    monkeypatch.setattr(click, 'confirm', monkey_patch_click)

    # Make sure that we are not ignoring warnings.
    checks.env.IGNORE_WARNINGS = False
    result = checks.WorkflowChecksMixin._check_if_uncommited_files()
    assert result is True

    _ignore = checks.env.IGNORE_WARNINGS
    checks.env.IGNORE_WARNINGS = True
    result = checks.WorkflowChecksMixin._check_if_uncommited_files()
    assert result is False

    checks.env.IGNORE_WARNINGS = _ignore


def test_check_if_remote_head_is_different(mocker, monkeypatch):
    """
    _check_if_remote_head_is_different() checks if remote SHA is
    different than local
    """
    def monkey_patch_git_command_fatal(*args, **kwargs):
        return 'fatal'

    monkeypatch.setattr(checks, 'execute_git_command', monkey_patch_git_command_fatal)
    assert checks.WorkflowChecksMixin._check_if_remote_head_is_different() is None

    def monkey_patch_git_command_same(*args, **kwargs):
        return 'foo'

    monkeypatch.setattr(checks, 'execute_git_command', monkey_patch_git_command_same)
    assert checks.WorkflowChecksMixin._check_if_remote_head_is_different() is False


def test_check_if_remote_head_is_different_shows_warning(mocker, monkeypatch):
    """
    _check_if_remote_head_is_different() shows warning if SHAs are different
    """
    def mp_git_command(*args, **kwargs):
        if r"@{u}" in args[0] or 'merge-base' in args[0]:
            return 'test-0'
        return 'test-1'

    def mp_confirm(*args, **kwargs):
        pass

    monkeypatch.setattr(click, 'confirm', mp_confirm)
    mocker.patch('click.confirm')
    monkeypatch.setattr(checks, 'execute_git_command', mp_git_command)

    result = checks.WorkflowChecksMixin._check_if_remote_head_is_different()
    assert result is True

    click.confirm.assert_called_once()


def test_check_github_repository(monkeypatch):
    """_check_github_repository() checks that we are in a git repository"""
    def mp_git_command_fatal(*args, **kwargs):
        return 'mygit.company.com'

    monkeypatch.setattr(checks, 'execute_git_command', mp_git_command_fatal)
    with pytest.raises(ClickException):
        checks.WorkflowChecksMixin._check_github_repository()

    def mp_git_command_fatal(*args, **kwargs):
        return None

    monkeypatch.setattr(checks, 'execute_git_command', mp_git_command_fatal)
    with pytest.raises(ClickException):
        checks.WorkflowChecksMixin._check_github_repository()



import ast
from typing import List

import click
from gql import gql
from rich.console import Console
from rich.table import Table
from yaspin import yaspin

import grid.globals as env

#  Maps backend class types to user-friendly messages.
TASK_CLASS_MAPPING = {
    'grid.core.repository_builder.RepositoryBuilder': 'Building container',
    'grid.core.cluster.Cluster': 'Creating cluster',
    'grid.core.trainer.experiment.Experiment': 'Scheduling experiment',
    'grid.core.trainer.run.RunNodePool': 'Creating node pool',
    'grid.core.trainer.interactive.InteractiveNodeTask': 'Creating interactive node',
    'grid.core.trainer.experiment.ExperimentsWarpSpeed': 'Scheduling experiment'
}

# Backend classes that we don't show to users. These
# are automatically triggered if a given user does not
# have global properties set.
TASK_CLASS_IGNORE = ('grid.core.clusters.deploy_tensorboard.ReconcileTensorboard',
                     'grid.core.clusters.global_user_cluster.ReconcileCluster', 'grid.core.user.ReconcileUser')


def style_status(format_string: str, status: str):
    """
    Styles a status message using click.stye.

    Parameters
    ----------
    status: str
        Status message to style.

    Return
    ------
    styled_status: str
        Styled string
    """
    styled_status = format_string

    if status == 'failed':
        styled_status = click.style(styled_status, fg='red')
    elif status in ('finished', 'ready'):
        styled_status = click.style(styled_status, fg='green')
    elif status in ('running', 'queued', 'pending'):
        styled_status = click.style(styled_status, fg='yellow')
    elif status in ('cancelled'):
        styled_status = click.style(styled_status, fg='white')

    return styled_status


class BaseObservable:
    def __init__(self, client, spinner_load_type=""):
        self.client = client
        self.console = Console()
        self.spinner = yaspin(text=f"Loading {spinner_load_type}...", color="yellow")

    @staticmethod
    def create_table(columns: List[str]) -> Table:
        table = Table(show_header=True, header_style="bold green")

        table.add_column(columns[0], style='dim')
        for column in columns[1:]:
            table.add_column(column, justify='right')

        return table

    def _get_task_run_dependencies(self, run_name: str):
        """Gets dependency data for a given Run"""
        query = gql("""
        query (
            $runName: ID!
        ) {
            getRunTaskStatus (
                runName: $runName
            ) {
                success
                runId
                name
                status
                message
                dependencies {
                    taskId
                    status
                    taskType
                    message
                    error
                }
            }
        }
        """)
        params = {'runName': run_name}

        #  Make GraphQL query.
        result = None
        try:
            result = self.client.execute(query, variable_values=params)
            if not result['getRunTaskStatus']['success']:
                raise Exception(result['getRunTaskStatus'])
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            self.spinner.fail("✘")
            self.spinner.stop()

            if env.DEBUG:
                click.echo(message)

            if 'not found' in message or 'No runs available' in message:
                raise click.ClickException(f'Run {run_name} not found. Did you cancel it already?')

        if result:
            dependencies = result['getRunTaskStatus']['dependencies']
            return dependencies

    def _get_task_run_status(self, run_name: str):
        #  Get dependency data.
        dependencies = self._get_task_run_dependencies(run_name=run_name)

        # short circuit if run has no dependencies
        if not dependencies:
            return

        #  Dict to collect all errors for given tasks.
        dependency_data = {
            k: {
                'statuses': [],
                'errors': [],
                'messages': []
            }
            for k in TASK_CLASS_MAPPING if k not in TASK_CLASS_IGNORE
        }
        for task in dependencies:
            if task['taskType'] in TASK_CLASS_IGNORE:
                continue
            dependency_data[task['taskType']]['statuses'].append(task['status'])
            dependency_data[task['taskType']]['errors'].append(task['error'])
            dependency_data[task['taskType']]['messages'].append(task['message'])

        #  Inform user that she can see error logs by passing
        #  a flag.
        all_statuses = []
        for status in dependency_data.values():
            all_statuses += status['statuses']

        if any(s == 'failed' for s in all_statuses) and \
            not env.SHOW_PROCESS_STATUS_DETAILS:
            click.echo(f'''
        The Run "{run_name}" failed to start due to a setup error.
        You can see the errors by running

            grid status {run_name} --details

            ''')

        #  If there's an error with and pre-run steps, then
        #  print that error to the terminal.
        if env.SHOW_PROCESS_STATUS_DETAILS:
            for key, value in dependency_data.items():
                for error, message, status in zip(value['errors'], value['messages'], all_statuses):
                    styled_key = style_status(TASK_CLASS_MAPPING[key], status)
                    if error:
                        for line in error.splitlines():
                            click.echo(f'[{styled_key}] {line}')
                    else:
                        click.echo(f'[{styled_key}] {message} ... ')



#!/usr/bin/env python
# Copyright 2020 Grid AI Inc.
"""Entrypoint for the Grid CLI."""
import sys
import traceback

import click
from click import ClickException

import grid.cli as cli
import grid.globals as env
from grid.metadata import __logo__
from grid.metadata import __version__
from grid.tracking import Segment
from grid.tracking import TrackingEvents
from grid.utilities import introspect_module


@click.group()
@click.option('--debug',
              type=bool,
              help='Used for logging additional information for debugging purposes.',
              is_flag=True)
def main(debug: bool = True) -> None:
    """Grid CLI"""
    if debug:
        env.logger.info('Starting gridrunner in DEGUB mode.')

    env.DEBUG = debug


@main.command()
def version() -> None:
    """
    Prints CLI version to stdout
    """
    logo = click.style(__logo__, fg='green')
    click.echo(logo)

    version = f"""
                                Grid CLI ({__version__})
                               https://docs.grid.ai
    """
    click.echo(version)


#  Adds all CLI commands. Commands are introspected
#  from the cli module.
for command in introspect_module(cli):
    command: click.Command
    main: click.Group
    main.add_command(command)

if __name__ == '__main__':
    tracker = Segment()
    tracker.send_event(event=TrackingEvents.CLI_STARTED)
    try:
        main.main(prog_name="grid", standalone_mode=False)
        tracker.send_event(event=TrackingEvents.CLI_FINISHED)
    except ClickException as e:
        tracker.report_exception(e)
        e.show()
        sys.exit(e.exit_code)
    except Exception as e:
        traceback.print_exc()
        tracker.report_exception(e)
        tracker.flush()
        sys.exit(1)
    finally:
        tracker.flush()



from grid.core.base import GridObject


def test_grid_object_update_meta(monkeypatch):
    class TestObject(GridObject):
        def __init__(self):
            # Prevents Grid from instantiating
            self._data = None
            pass

    G = TestObject()
    previous = dir(G)

    G._update_meta()
    not_changed = dir(G)

    assert previous == not_changed

    # Assert dict keys are added as attributes
    G._data = {"test": True}
    G._update_meta()
    changed = dir(G)
    assert previous != changed
    assert G.test



import click
import pytest

from grid.cli.grid_train import _check_is_valid_extension
from grid.cli.grid_train import _check_run_name_is_valid
from grid.cli.grid_train import _resolve_instance_type_nickname
from grid.utilities import check_description_isnt_too_long


class TestTrainCallbacks:
    """Tests callbacks in train."""
    def test_name_callback(self):
        # Run name needs to be alphanumeric
        with pytest.raises(click.ClickException):
            _check_run_name_is_valid(None, None, '$')

        # Run name can't start or end with a dash
        with pytest.raises(click.ClickException):
            _check_run_name_is_valid(None, None, '-run')

        with pytest.raises(click.ClickException):
            _check_run_name_is_valid(None, None, 'run-')

        # Run name can't contain upper case letters
        with pytest.raises(click.ClickException):
            _check_run_name_is_valid(None, None, 'Run')

        # Besides that, run name should work
        assert _check_run_name_is_valid(None, None, 'run')

    def test_nickname(self):
        aws_node_to_nicknames = {
            'p3.16xlarge': '8_v100_16gb',
            'p3dn.24xlarge': '8_v100_32gb',
            'g4dn.metal': '8_t4_16gb',
            'p2.8xlarge': '8_k80_12gb',
            'p3.8xlarge': '4_v100_16gb',
            'g4dn.12xlarge': '4_t4_16gb',
            'g3.16xlarge': '4_m60_8gb',
            'g3.8xlarge': '2_m60_8gb',
            'p3.2xlarge': '1_v100_16gb',
            'p3.2xlarge': '1_V100_16Gb',
            'g4dn.16xlarge': '1_t4_16gb',
            'p2.xlarge': '1_k80_12gb',
            'g3.4xlarge': '1_m60_8gb',
            't2.large': '2_cpu_8gb',
            't2.large': '2_CPU_8GB',
            't2.medium': '2_cpu_4gb'
        }
        for k, v in aws_node_to_nicknames.items():

            assert _resolve_instance_type_nickname(None, None, v) == k

    def test_valid_extension_callback(self):
        """
        _check_is_valid_extension() raises exception if file isn't valid extension
        """
        f = 'foo.txt'
        with pytest.raises(click.BadParameter):
            _check_is_valid_extension(None, None, f)

    def test_check_description_isnt_too_long(self):
        """
        _check_description_isnt_too_long() checks that description isn't
        too long
        """
        d = 'a' * 199
        assert d == check_description_isnt_too_long(None, None, d)

        d = 'a' * 201
        with pytest.raises(click.BadParameter):
            check_description_isnt_too_long(None, None, d)



import click
import pytest
from tests.utilities import captured_output

from grid.cli.grid_credentials import print_creds_table


class TestCredentialsUtilities:
    """Test utilities in credentials."""
    @staticmethod
    def test_print_creds_table_raises_exception():
        with pytest.raises(click.ClickException):
            print_creds_table(creds=[])

    @staticmethod
    def test_print_creds_table_prints_table():
        credentials = [{
            'credentialId': 'test',
            'provider': 'test',
            'alias': 'test',
            'createdAt': '2020-05-16T21:21:10.745898+00:00',
            'lastUsedAt': 'test',
            'defaultCredential': 'test'
        }]
        expected_columns = ['Credential', 'Provider', 'Alias', 'Created At', 'Last Used At', 'Default']

        with captured_output() as (out, _):
            print_creds_table(creds=credentials)

        output = out.getvalue().strip()
        for column in expected_columns:
            assert column in output



from typing import Optional

import click

from grid.cli.grid_train import get_credentials
from grid.client import Grid


@click.group(invoke_without_command=True)
@click.pass_context
def datastores(ctx) -> None:
    """Manages datastore workflows in Grid"""
    client = Grid()

    if ctx.invoked_subcommand is None:
        client.list_datastores()


@datastores.command()
@click.argument('session_name', nargs=1)
@click.pass_context
def resume(ctx, session_name: str):
    """Resumes uploading a datastore"""
    client = Grid()
    if session_name == "list":
        client.list_resumable_datastore_sessions()
        return

    client.resume_datastore_session(session_name)


@datastores.command()
@click.option('--source',
              required=True,
              help="""
              Source to create datastore from. This could either be a local directory
              (e.g: /opt/local_folder) or a remote HTTP URL pointing to a TAR file
              (e.g: http://some_domain/data.tar.gz)
              """)
@click.option('--name', type=str, required=True, help='Name of the datastore')
@click.option('--grid_credential', type=str, required=False, help='Grid credential ID that will store this datastore')
@click.option('--compression',
              type=bool,
              required=False,
              help='Compresses datastores with GZIP when flag is passed.',
              default=False,
              is_flag=True)
@click.pass_context
def create(ctx, source: str, name: str, grid_credential: Optional[str] = None, compression: bool = False) -> None:
    """Creates datastores"""
    client = Grid()

    credential = get_credentials(client=client, grid_credential=grid_credential)

    client.upload_datastore(source=source, credential_id=credential['credentialId'], name=name, compression=compression)


@datastores.command()
@click.pass_context
def list(ctx) -> None:
    """Lists datastores"""
    client = Grid()
    client.list_datastores()


@datastores.command()
@click.option('--name', type=str, required=True, help='Name of the datastore')
@click.option('--version', type=int, required=True, help='Version of the datastore')
@click.option('--grid_credential', type=str, required=True, help='Credential Id for this datastore')
@click.pass_context
def delete(ctx, name: str, version: int, grid_credential: str) -> None:
    """Lists datastores"""
    client = Grid()
    client.delete_datastore(name=name, version=version, credential_id=grid_credential)



from click.testing import CliRunner
from requests.exceptions import HTTPError
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
from grid.cli.grid_login import click
from grid.cli.grid_login import env
import grid.client as grid
import grid.commands.credentials as credentials

RUNNER = CliRunner()


class TestCredentials:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client

        credentials.gql = lambda x: x
        grid.gql = lambda x: x

        create_test_credentials()

    def test_login_without_username_and_api_key(self, monkeypatch):
        """
        grid login without username and API key creates a prompt for username
        """
        def mp_launch(*args, **kwargs):
            return True

        monkeypatch.setattr(click, 'launch', mp_launch)
        result = RUNNER.invoke(cli.login, [], input='testuser\n apikey\n')
        assert result.exit_code == 0
        assert not result.exception
        assert 'username' in result.output
        assert 'Login successful' in result.output

    def test_login_with_username_skips_prompt(self, monkeypatch):
        """
        grid login with --username skips prompt
        """
        def mp_launch(*args, **kwargs):
            return True

        monkeypatch.setattr(click, 'launch', mp_launch)
        result = RUNNER.invoke(cli.login, ['--username', 'testuser'], input='apikey\n')
        assert result.exit_code == 0
        assert not result.exception
        assert 'username' not in result.output
        assert 'Login successful' in result.output

    def test_login_fails_with_auth_error(self, monkeypatch):
        """grid login raises error when auth is wrong"""
        def mp_launch(*args, **kwargs):
            return True

        # Not authorized headers.
        def mp_init_client(self, *args, **kwargs):
            class A:
                @staticmethod
                def execute(*args, **kwargs):
                    raise HTTPError('not authorized')

            self.client = A

        monkeypatch.setattr(grid.Grid, '_init_client', mp_init_client)
        monkeypatch.setattr(click, 'launch', mp_launch)
        result = RUNNER.invoke(cli.login, ['--username', 'testuser'], input='apikey\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to login' in result.output

        # API error.
        def mp_cli_login(*args, **kwargs):
            raise Exception('failed to make query')

        monkeypatch.setattr(resolvers, 'cli_login', mp_cli_login)
        monkeypatch.setattr(click, 'launch', mp_launch)
        env.DEBUG = True
        result = RUNNER.invoke(cli.login, ['--username', 'testuser'], input='apikey\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to login' in result.output



import click
import pytest
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client
import websockets

import grid.client as grid
import grid.observables.base as base_observable
import grid.observables.experiment as experiment_observable


def mp_execute(*args, **kwargs):
    raise Exception("{'message':'not found'}")


class TestExperimentObservable:
    @classmethod
    def setup_class(cls):
        create_test_credentials()

        grid.Grid._init_client = monkey_patch_client

        experiment_observable.gql = lambda x: x
        base_observable.gql = lambda x: x

        cls.grid = grid.Grid(load_local_credentials=False)
        cls.grid._init_client()

    def test_get(self, capsys):
        obsevable = experiment_observable.Experiment(client=self.grid.client, identifier='test-experiment')
        result = obsevable.get()
        assert len(result) > 0

        captured = capsys.readouterr()
        assert 'running' in captured.out

    def test_get_history(self, capsys):
        obsevable = experiment_observable.Experiment(client=self.grid.client, identifier='test-run')
        obsevable.get_history()

        captured = capsys.readouterr()
        assert 'failed' in captured.out

    def test_follow(self, capsys):
        obsevable = experiment_observable.Experiment(client=self.grid.client, identifier='test-run')
        obsevable.follow()

        captured = capsys.readouterr()
        assert 'Following Run details' in captured.out

    def test_follow_handles_exceptions(self, capsys, monkeypatch):
        def mp_websocket_closed(*args, **kwargs):
            raise websockets.exceptions.ConnectionClosedError(code=1, reason='test')

        monkeypatch.setattr(self.grid.client, 'subscribe', mp_websocket_closed)
        obsevable = experiment_observable.Experiment(client=self.grid.client, identifier='test-run')
        with pytest.raises(click.ClickException):
            obsevable.follow()

        # def mp_websocket_connection_closed_ok(*args, **kwargs):
        #     raise websockets.exceptions.ConnectionClosedOK(code=1, reason='test')

        # monkeypatch.setattr(self.grid.client, 'subscribe', mp_websocket_connection_closed_ok)
        # obsevable = observables.Experiment(client=self.grid.client,
        #                                 identifier='test-run')
        # with pytest.raises(click.ClickException):
        #     obsevable.follow()

        # def mp_execute(*args, **kwargs):
        #     raise Exception("Server error: {'message':'not found'}")

        # monkeypatch.setattr(self.grid.client, 'subscribe', mp_execute)
        # obsevable = observables.Experiment(client=self.grid.client,
        #                                 identifier='test-run')
        # with pytest.raises(click.ClickException):
        #     obsevable.follow()

    def test_render_no_expriment(self):
        obsevable = experiment_observable.Experiment(client=self.grid.client, identifier='test-run')
        table = obsevable.render_experiments([])
        assert table.row_count == 0
        assert [x.header for x in table.columns] == ["Experiment", "Command", "Status", "Duration"]

    def test_render_no_hparam_expriment(self):
        obsevable = experiment_observable.Experiment(client=self.grid.client, identifier='test-run')
        table = obsevable.render_experiments([{
            'experimentId': 'test',
            'name': 'gentle-grasshopper-960-exp0',
            'status': 'succeeded',
            'invocationCommands': 'python pl_mnist.py',
            'createdAt': '2020-11-16T21:01:29.813309+00:00',
            'finishedAt': '2020-11-16T21:44:05+00:00',
            'commitSha': 'e2f16c5c14d52c332568718bfef9f632e3191825',
            'run': {
                'runId': '34dfb18f-d64b-4a67-ac68-773022a81996'
            },
            'startedRunningAt': '2020-11-16T21:18:45+00:00'
        }])
        assert table.row_count == 1
        assert [x.header for x in table.columns] == ["Experiment", "Command", "Status", "Duration"]
        assert [x._cells[0] for x in table.columns] == [
            'gentle-grasshopper-960-exp0',
            'pl_mnist.py',
            'succeeded',
            '0d-00:25:20',
        ]

    def test_render_hparam_expriment(self):
        obsevable = experiment_observable.Experiment(client=self.grid.client, identifier='test-run')
        table = obsevable.render_experiments([{
            'experimentId': 'test',
            'name': 'gentle-grasshopper-960-exp0',
            'status': 'succeeded',
            'invocationCommands': 'python pl_mnist.py --lr 0.5',
            'createdAt': '2020-11-16T21:01:29.813309+00:00',
            'finishedAt': '2020-11-16T21:44:05+00:00',
            'commitSha': 'e2f16c5c14d52c332568718bfef9f632e3191825',
            'run': {
                'runId': '34dfb18f-d64b-4a67-ac68-773022a81996'
            },
            'startedRunningAt': '2020-11-16T21:18:45+00:00'
        }])
        assert table.row_count == 1
        assert [x.header for x in table.columns] == ["Experiment", "Command", "Status", "Duration", "lr"]
        assert [x._cells[0] for x in table.columns] == [
            'gentle-grasshopper-960-exp0',
            'pl_mnist.py',
            'succeeded',
            '0d-00:25:20',
            '0.5',
        ]
